import asyncio
from typing import Dict, Optional

from bxcommon.rpc.json_rpc_response import JsonRpcResponse


class ResponseQueue:
    """
    Queue for matching RPC requests with responses.
    Usage of this class expects all RPC requests to contain request IDs.

    This class doesn't clean up its memory usage, so if the corresponding RPC
    server doesn't respond to request IDs this queue will continue growing.
    """
    def __init__(self, only_once: bool = True):
        self.only_once = only_once
        self.message_notifiers: Dict[str, asyncio.Event] = {}
        self.message_by_request_id: Dict[str, JsonRpcResponse] = {}

    async def put(self, message: JsonRpcResponse) -> None:
        request_id = message.id
        if request_id is None:
            raise ValueError(
                "Response queue cannot accept RPC messages with no request ID."
            )
        self.message_by_request_id[request_id] = message

        if request_id in self.message_notifiers:
            self.message_notifiers[request_id].set()

    async def get_by_request_id(self, request_id: str) -> Optional[JsonRpcResponse]:
        if request_id in self.message_by_request_id:
            message = self.message_by_request_id[request_id]
            if self.only_once:
                self.cleanup(request_id)
            return message

        if request_id in self.message_notifiers:
            event = self.message_notifiers[request_id]
        else:
            event = asyncio.Event()
            self.message_notifiers[request_id] = event

        await event.wait()

        if request_id not in self.message_by_request_id:
            return None
        message = self.message_by_request_id[request_id]

        if self.only_once:
            self.cleanup(request_id)

        return message

    def cleanup(self, request_id: str) -> None:
        self.message_notifiers.pop(request_id, None)
        self.message_by_request_id.pop(request_id, None)



from typing import TYPE_CHECKING

from bxcommon.feed.feed import FeedKey
from bxcommon.utils.blockchain_utils.eth import eth_common_utils
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.connections.abstract_relay_connection import AbstractRelayConnection
from bxcommon.feed.eth.eth_new_transaction_feed import EthNewTransactionFeed
from bxcommon.feed.eth.eth_raw_transaction import EthRawTransaction
from bxcommon.feed.feed_source import FeedSource

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode


class EthRelayConnection(AbstractRelayConnection):
    node: "EthGatewayNode"

    def publish_new_transaction(
        self, tx_hash: Sha256Hash, tx_contents: memoryview, local_region: bool
    ) -> None:
        gas_price = eth_common_utils.raw_tx_gas_price(tx_contents, 0)
        if gas_price >= self.node.get_network_min_transaction_fee():
            self.node.feed_manager.publish_to_feed(
                FeedKey(EthNewTransactionFeed.NAME, network_num=self.node.network_num),
                EthRawTransaction(
                    tx_hash, tx_contents, FeedSource.BDN_SOCKET, local_region=local_region
                )
            )



from bxutils.log_message_categories import *
from bxutils.logging_messages_utils import LogMessage

EMPTY_BLOCKCHAIN_NETWORK_LIST = LogMessage(
    "C-000000",
    REQUEST_RESPONSE_CATEGORY,
    "Empty list for blockchain networks from SDN, trying to obtain info from cache"
)
READ_CACHE_FILE_WARNING = LogMessage(
    "C-000001",
    GENERAL_CATEGORY,
    "Unable to find cache file: {}. Not using cache. You can disable caching by setting --enable-node-cache False."
)
PROTOCOL_VERSION_NOT_IN_FACTORY_MAPPING = LogMessage(
    "C-000002",
    REQUEST_RESPONSE_CATEGORY,
    "Got a message with version {}. Should be supported, but not in factory mapping."
)
UNABLE_TO_DETERMINE_PUBLIC_IP = LogMessage(
    "C-000003",
    NETWORK_CATEGORY,
    ("Unable to determine public IP address, please specify one manually via the '--external-ip' command line "
     "argument.\n\n Detailed error message:\n\t{}")
)
PING_TRIGGERED_AN_ERROR = LogMessage(
    "C-000004",
    NETWORK_CATEGORY,
    "Ping to {} {} triggered an error: {}."
)
ERROR_LOADING_MODEL_INTO_DICT = LogMessage(
    "C-000005",
    GENERAL_CATEGORY,
    "Failed when tried to load str to dict. model class: {} model params: {}"
)
STOP_RECORDING_CALLED_ON_UNINITIALIZED_THREAD = LogMessage(
    "C-000006",
    LOGGING_CATEGORY,
    "Thread was not initialized yet, but stop_recording was called. An invariant in the code is broken."
)
FAILURE_RECORDING_STATS = LogMessage(
    "C-000007",
    LOGGING_CATEGORY,
    "Recording {} stats failed with exception: {}. Stack trace: {}"
)
HTTP_REQUEST_RETURNED_ERROR = LogMessage(
    "C-000008",
    REQUEST_RESPONSE_CATEGORY,
    "{} to {} returned error: {}."
)
TASK_CANCELLED = LogMessage(
    "C-000009",
    THREAD_CATEGORY,
    "Task: {} with values: {} was cancelled: {}"
)
TASK_FAILED = LogMessage(
    "C-000010",
    THREAD_CATEGORY,
    "Task: {} with values: {} failed due to error: {}"
)
FLUENTD_LOGGER_BUFFER_OVERFLOW = LogMessage(
    "C-000011",
    LOGGING_CATEGORY,
    "fluentd logger, buffer overflow"
)
INVALID_LOG_LEVEL = LogMessage(
    "C-000012",
    LOGGING_CATEGORY,
    "Invalid Log Level Provided Ignore for path {}: {}"
)
DETECTED_PYTHON3_6 = LogMessage(
    "C-000013",
    GENERAL_CATEGORY,
    ("Python 3.6 environment is detected. Degraded performance is expected. Upgrade "
     "to Python 3.7 or above for improved performance.")
)
EMPTY_BLOCKCHAIN_NETWORK_CACHE = LogMessage(
    "C-000014",
    PROCESSING_FAILED_CATEGORY,
    "Cached info for blockchain_networks was empty"
)
DECRYPTION_FAILED = LogMessage(
    "C-000015",
    PROCESSING_FAILED_CATEGORY,
    "Could not decrypt encrypted item with hash {}. Last four bytes: {}"
)
UNABLE_TO_DETERMINE_CONNECTION_TYPE = LogMessage(
    "C-000016",
    CONNECTION_PROBLEM_CATEGORY,
    "Could not determine expected connection type for {}:{}. Disconnecting..."
)
FAILED_TO_AUTHENTICATE_CONNECTION = LogMessage(
    "C-000017",
    AUTHENTICATION_ERROR,
    "Failed to authenticate connection on {}:{} due to an error: {}."
)
ATTEMPTED_TO_ASSIGN_NULL_SHORT_ID_TO_TX_HASH = LogMessage(
    "C-000018",
    PROCESSING_FAILED_CATEGORY,
    "Attempted to assign null short id to transaction hash {}. Ignoring."
)
SID_MEMORY_MANAGEMENT_FAILURE = LogMessage(
    "C-000019",
    MEMORY_CATEGORY,
    "Memory management failure. There appears to be a lack of short ids in the node. Clearing all transaction data: {}"
)
UNABLE_TO_DETERMINE_TX_FINAL_CONFIRMATIONS_COUNT = LogMessage(
    "C-000020",
    PROCESSING_FAILED_CATEGORY,
    "Could not determine final confirmations count for network number {}. Using default {}."
)
UNABLE_TO_DETERMINE_TX_EXPIRATION_TIME = LogMessage(
    "C-000021",
    PROCESSING_FAILED_CATEGORY,
    "Could not determine expiration time for transaction removed from cache for network number {}. Using default {}."
)
TX_CACHE_SIZE_LIMIT_NOT_CONFIGURED = LogMessage(
    "C-000022",
    PROCESSING_FAILED_CATEGORY,
    "Blockchain network {} does not have tx cache size limit configured. Using default {}."
)
UNABLE_TO_DETERMINE_TX_MEMORY_LIMIT = LogMessage(
    "C-000023",
    MEMORY_CATEGORY,
    "Could not determine transactions memory limit for network number {}. Using default {}."
)
THREADED_REQUEST_HAS_LONG_RUNTIME = LogMessage(
    "C-000024",
    PROCESSING_FAILED_CATEGORY,
    "Threaded request was enqueued more than {} second(s) ago and hasn't finished yet: {}"
)
THREADED_REQUEST_IS_STALE = LogMessage(
    "C-000025",
    PROCESSING_FAILED_CATEGORY,
    "Threaded request hasn't started running yet, cancelling: {}"
)
BDN_RETURNED_NO_PEERS = LogMessage(
    "C-000026",
    REQUEST_RESPONSE_CATEGORY,
    "BDN returned no peers at endpoint: {}"
)
BDN_RETURNED_UNEXPECTED_NUMBER_OF_PEERS = LogMessage(
    "C-000027",
    REQUEST_RESPONSE_CATEGORY,
    "BDN did not send the expected number of remote blockchain peers."
)
BDN_CONTAINS_NO_CONFIGURED_NETWORKS = LogMessage(
    "C-000028",
    REQUEST_RESPONSE_CATEGORY,
    "BDN does not seem to contain any configured networks."
)
COULD_NOT_PARSE_MESSAGE = LogMessage(
    "C-000029",
    PROCESSING_FAILED_CATEGORY,
    "Could not parse message. Error: {}"
)
OUT_OF_MEMORY = LogMessage(
    "C-000030",
    PROCESSING_FAILED_CATEGORY,
    "Out of memory error occurred during message processing. Error: {}. "
)
UNAUTHORIZED_MESSAGE = LogMessage(
    "C-000031",
    PROCESSING_FAILED_CATEGORY,
    "Unauthorized message {} from {}."
)
UNABLE_TO_RECOVER_PARTIAL_MESSAGE = LogMessage(
    "C-000032",
    PROCESSING_FAILED_CATEGORY,
    "Unable to recover after message that failed validation. Closing connection."
)
TRYING_TO_RECOVER_MESSAGE = LogMessage(
    "C-000033",
    PROCESSING_FAILED_CATEGORY,
    "Message processing error; trying to recover. Error: {}."
)
UNABLE_TO_RECOVER_FULL_MESSAGE = LogMessage(
    "C-000034",
    PROCESSING_FAILED_CATEGORY,
    "Message processing error; unable to recover. Error: {}."
)
UNEXPECTED_MESSAGE = LogMessage(
    "C-000035",
    PROCESSING_FAILED_CATEGORY,
    "Received unexpected message ({}) before handshake completed. Closing."
)
MESSAGE_VALIDATION_FAILED = LogMessage(
    "C-000036",
    PROCESSING_FAILED_CATEGORY,
    "Message validation failed for {} message: {}."
)
INVALID_HANDSHAKE = LogMessage(
    "C-000037",
    PROCESSING_FAILED_CATEGORY,
    "Invalid handshake request on {}:{}. Rejecting the connection. {}"
)
DUPLICATE_CONNECTION = LogMessage(
    "C-000038",
    PROCESSING_FAILED_CATEGORY,
    "Discovered duplicate connections to node {}: {}. Closing."
)
TOO_MANY_BAD_MESSAGES = LogMessage(
    "C-000039",
    PROCESSING_FAILED_CATEGORY,
    "Received too many bad messages. Closing."
)
NETWORK_NUMBER_MISMATCH = LogMessage(
    "C-000040",
    PROCESSING_FAILED_CATEGORY,
    "Network number mismatch. Current network num {}, remote network num {}. Closing connection."
)
RPC_COULD_NOT_PARSE_TRANSACTION = LogMessage(
    "C-000041",
    REQUEST_RESPONSE_CATEGORY,
    "Error parsing the transaction:\n{}",
)
INTERNAL_ERROR_HANDLING_RPC_REQUEST = LogMessage(
    "C-000042",
    RPC_ERROR,
    "Internal error {} while handling request {} from {}"
)
BDN_ACCOUNT_INFORMATION_TIMEOUT = LogMessage(
    "C-000043",
    CONNECTION_PROBLEM_CATEGORY,
    "BDN does not send the account information within {} seconds."
)

NODE_RECEIVED_TX_WITH_INVALID_FORMAT = LogMessage(
    "C-000044",
    GENERAL_CATEGORY,
    "{} received transaction {} from {} with an invalid format"
)

NODE_RECEIVED_TX_WITH_INVALID_SIG = LogMessage(
    "C-000045",
    GENERAL_WARNING,
    "{} received transaction {} from {} with an invalid signature"
)
MISSING_ASSIGN_TIME_FOR_SHORT_ID = LogMessage(
    "C-000046",
    GENERAL_CATEGORY,
    "Missing assignment time for short id {}."
)
WS_COULD_NOT_CONNECT = LogMessage(
    "C-000047",
    CONNECTION_PROBLEM_CATEGORY,
    "Could not connect to websockets server at {}. Connection timed out.",
)
ETH_RPC_ERROR = LogMessage(
    "C-000048",
    PROCESSING_FAILED_CATEGORY,
    "RPC Error response: {}. details: {}.",
)
ETH_RPC_PROCESSING_ERROR = LogMessage(
    "C-000049",
    PROCESSING_FAILED_CATEGORY,
    "Encountered exception when processing message: {}. Error: {}. Continuing processing.",
)
ETH_WS_SUBSCRIBER_CONNECTION_BROKEN = LogMessage(
    "C-000050",
    RPC_ERROR,
    "Ethereum websockets connection was broken. Attempting reconnection..."
)
ETH_RPC_COULD_NOT_RECONNECT = LogMessage(
    "C-000051",
    CONNECTION_PROBLEM_CATEGORY,
    "Could not reconnect to Ethereum websockets feed. Disabling Ethereum transaction "
    "verification for now, but will attempt reconnection upon when the next subscriber "
    "reconnects."
)
MSG_PROXY_REQUESTER_QUEUE_EMPTY_ON_RESPONSE = LogMessage(
    "C-000052",
    GENERAL_CATEGORY,
    "Message proxy requester queue empty upon receiving response from remote blockchain "
    "node. Unable to forward response to blockchain node."
)
RPC_TRANSPORT_EXCEPTION = LogMessage(
    "C-000053",
    CONNECTION_PROBLEM_CATEGORY,
    "Websocket disconnected due to transport layer exception: {}",
)
BAD_RPC_SUBSCRIBER = LogMessage(
    "C-000054",
    RPC_ERROR,
    "Subscription message queue was completed filled up (size {}). "
    "Closing subscription RPC handler and all related subscriptions: {}"
)
COULD_NOT_SERIALIZE_FEED_ENTRY = LogMessage(
    "C-000055",
    PROCESSING_FAILED_CATEGORY,
    "Could not serialize feed entry. Skipping."
)
BAD_FEED_SUBSCRIBER = LogMessage(
    "C-000056",
    GENERAL_CATEGORY,
    "Subscriber {} was not receiving messages and emptying its queue from "
    "{}. Disconnecting.",
)
WS_COULD_NOT_CONNECT_AFTER_RETRIES = LogMessage(
    "C-000057",
    CONNECTION_PROBLEM_CATEGORY,
    "Could not connect to websockets server at {}. Connection timed out. Tried {} times.",
)
WS_UNEXPECTED_ERROR = LogMessage(
    "C-000058",
    CONNECTION_PROBLEM_CATEGORY,
    "Unexpected logger connection error: {}. Retrying connection."
)
WS_COULD_NOT_PROCESS_NOTIFICATION = LogMessage(
    "C-000059",
    PROCESSING_FAILED_CATEGORY,
    "Unexpected error in feed callback: {}."
)



from typing import List

from bxcommon.utils.blockchain_utils.btc.btc_common_utils import btc_varint_to_int
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash
from bxgateway.btc_constants import BTC_HDR_COMMON_OFF, BTC_SHA_HASH_LEN, BTC_VARINT_MIN_SIZE
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.btc.btc_messages_util import pack_int_to_btc_varint, get_next_tx_size


class BlockTransactionsBtcMessage(BtcMessage):
    """
    Message used as reply to Get Block Transactions message
    """

    MESSAGE_TYPE = BtcMessageType.BLOCK_TRANSACTIONS

    # pyre-fixme[9]: magic has type `int`; used as `None`.
    # pyre-fixme[9]: block_hash has type `BtcObjectHash`; used as `None`.
    # pyre-fixme[9]: transactions has type `List[memoryview]`; used as `None`.
    def __init__(self, magic: int = None, block_hash: BtcObjectHash = None, transactions: List[memoryview] = None,
                 # pyre-fixme[9]: buf has type `memoryview`; used as `None`.
                 buf: memoryview = None):
        if buf is None:
            total_tx_size = sum(len(tx) for tx in transactions)
            # pyre-fixme[9]: buf has type `memoryview`; used as `bytearray`.
            buf = bytearray(BTC_HDR_COMMON_OFF + BTC_SHA_HASH_LEN + BTC_VARINT_MIN_SIZE + total_tx_size)
            off = BTC_HDR_COMMON_OFF

            buf[off:off + BTC_SHA_HASH_LEN] = block_hash.get_big_endian()
            off += BTC_SHA_HASH_LEN

            off += pack_int_to_btc_varint(len(transactions), buf, off)

            for tx in transactions:
                buf[off:off + len(tx)] = tx
                off += len(tx)

            self.buf = buf
            super(BlockTransactionsBtcMessage, self).__init__(magic, self.MESSAGE_TYPE, off - BTC_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._block_hash = None
        # pyre-fixme[8]: Attribute has type `List[int]`; used as `None`.
        self._transactions: List[int] = None

    def block_hash(self) -> BtcObjectHash:
        if self._block_hash is None:
            self._block_hash = BtcObjectHash(buf=self.buf, offset=BTC_HDR_COMMON_OFF, length=BTC_SHA_HASH_LEN)

        # pyre-fixme[7]: Expected `BtcObjectHash` but got `None`.
        return self._block_hash

    def transactions(self) -> List[memoryview]:
        if self._transactions is None:
            self._transactions = []
            off = BTC_HDR_COMMON_OFF + BTC_SHA_HASH_LEN

            tx_len, size = btc_varint_to_int(self.buf, off)
            off += size

            for _ in range(tx_len):
                size = get_next_tx_size(self.buf, off)
                tx = self._memoryview[off:off + size]
                off += size
                self._transactions.append(tx)

        return self._transactions  # pyre-ignore



from bxcommon.utils.flag_enum import Flag, FlagCollection


class SocketConnectionState(Flag):
    pass


class SocketConnectionStates(FlagCollection):
    CONNECTING = SocketConnectionState()
    INITIALIZED = SocketConnectionState()
    DO_NOT_RETRY = SocketConnectionState()
    HALT_RECEIVE = SocketConnectionState()


SocketConnectionStates.init(SocketConnectionState)



import struct

from bxcommon import constants
from bxcommon.exceptions import PayloadLenError
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxutils import logging

logger = logging.get_logger(__name__)


class AbstractInternalMessage(AbstractMessage):
    MESSAGE_TYPE = BloxrouteMessageType.ABSTRACT_INTERNAL
    HEADER_LENGTH = constants.BX_HDR_COMMON_OFF
    STARTING_BYTES_LEN = 0

    def __init__(self, msg_type: bytes, payload_len: int, buf: bytearray) -> None:

        if len(buf) < self.HEADER_LENGTH:
            raise ValueError("Buffer must be at least {0} in length.".format(self.HEADER_LENGTH))

        if payload_len < 0:
            raise ValueError("Payload_len must be a positive integer.")

        off = self.STARTING_BYTES_LEN
        struct.pack_into("<12sL", buf, off, msg_type, payload_len)
        off += constants.BX_HDR_COMMON_OFF

        self.buf = buf
        self._memoryview = memoryview(buf)

        self._msg_type = msg_type
        self._payload_len = payload_len
        self._payload = None
        self._priority = False

    # TODO: pull this out in to a message protocol
    @classmethod
    def unpack(cls, buf):
        command, payload_length = struct.unpack_from("<12sL", buf, cls.STARTING_BYTES_LEN)
        return command.rstrip(constants.MSG_NULL_BYTE), payload_length

    @classmethod
    def validate_payload(cls, buf, unpacked_args) -> None:
        _command, payload_length = unpacked_args
        if payload_length != len(buf) - cls.HEADER_LENGTH:
            error_message = "Payload length does not match buffer size: {} vs {} bytes" \
                .format(payload_length, len(buf) - cls.HEADER_LENGTH)
            logger.error(error_message)
            raise PayloadLenError(error_message)

    # pylint: disable=protected-access
    @classmethod
    def initialize_class(cls, cls_type, buf, unpacked_args):
        command, payload_length = unpacked_args
        instance = cls_type(buf=buf)
        instance._msg_type = command
        instance._payload_len = payload_length
        return instance

    # END TODO

    def rawbytes(self) -> memoryview:
        """
        Returns a memoryview of the message
        """
        if self._payload_len is None:
            self._payload_len, = struct.unpack_from("<L", self.buf, self.STARTING_BYTES_LEN + constants.MSG_TYPE_LEN)

        if self._payload_len + self.HEADER_LENGTH == len(self.buf):
            return self._memoryview
        else:
            return self._memoryview[0:self._payload_len + self.HEADER_LENGTH]

    def msg_type(self) -> bytes:
        if self._msg_type is None:
            self._msg_type = struct.unpack_from("<12s", self.buf, self.STARTING_BYTES_LEN)[0]
        return self._msg_type

    def payload_len(self) -> int:
        if self._payload_len is None:
            self._payload_len = struct.unpack_from("<L", self.buf, self.STARTING_BYTES_LEN + constants.MSG_TYPE_LEN)[0]
        return self._payload_len

    def payload(self):
        if self._payload is None:
            self._payload = self.buf[self.HEADER_LENGTH:self.payload_len() + self.HEADER_LENGTH]
        return self._payload

    def set_priority(self, high: bool) -> None:
        """
        Sets the priority of this message (to be high or low)
        :param high: True if this is high priority. False otherwise.
        """
        self._priority = high

    def get_priority(self) -> bool:
        return self._priority

    def __eq__(self, other) -> bool:
        """
        Expensive equality comparison. Use only for tests.
        """
        if not isinstance(other, AbstractInternalMessage):
            return False
        else:
            return self.rawbytes().tobytes() == other.rawbytes().tobytes()

    def __repr__(self) -> str:
        return "Message<type: {}, length: {}>".format(self.MESSAGE_TYPE, len(self.rawbytes()))



from typing import TYPE_CHECKING

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxgateway.connections.eth.eth_base_connection import EthBaseConnection
from bxgateway.connections.eth.eth_remote_connection_protocol import EthRemoteConnectionProtocol

if TYPE_CHECKING:
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode


class EthRemoteConnection(EthBaseConnection):
    CONNECTION_TYPE = ConnectionType.REMOTE_BLOCKCHAIN_NODE

    def __init__(self, sock: AbstractSocketConnectionProtocol, node: "EthGatewayNode"):
        super(EthRemoteConnection, self).__init__(sock, node)
        self.connection_protocol = EthRemoteConnectionProtocol(
            self, self.is_handshake_initiator, self.rlpx_cipher
        )

    def connection_public_key(
        self, _sock: AbstractSocketConnectionProtocol, node: "EthGatewayNode"
    ) -> bytes:
        return node.get_remote_public_key()



from hashlib import sha256

from nacl.secret import SecretBox
from nacl.utils import random

# Length of a SHA256 double hash
SHA256_HASH_LEN = 32
KEY_SIZE = SecretBox.KEY_SIZE


def bitcoin_hash(content):
    return sha256(sha256(content).digest()).digest()


def double_sha256(content):
    return sha256(sha256(content).digest()).digest()


def symmetric_encrypt(content, key=None):
    if not key:
        key = random(KEY_SIZE)
    ciphertext = SecretBox(key).encrypt(content)
    return key, ciphertext


def symmetric_decrypt(key, ciphertext):
    return SecretBox(key).decrypt(ciphertext)



import base64
import json
import struct
import time
from collections import deque
from datetime import datetime
from typing import Tuple, List, Deque, Union

from bxutils import logging
from bxutils.encoding.json_encoder import EnhancedJSONEncoder

from bxcommon import constants
from bxcommon.messages.bloxroute import compact_block_short_ids_serializer
from bxcommon.messages.bloxroute.compact_block_short_ids_serializer import BlockOffsets
from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils import convert, crypto
from bxcommon.utils.blockchain_utils.ont.ont_object_hash import OntObjectHash
from bxcommon.utils.object_hash import Sha256Hash

from bxgateway import ont_constants, log_messages
from bxgateway.abstract_message_converter import BlockDecompressionResult, finalize_block_bytes
from bxgateway.messages.ont import ont_messages_util
from bxgateway.messages.ont.abstract_ont_message_converter import AbstractOntMessageConverter, get_block_info
from bxgateway.messages.ont.consensus_ont_message import OntConsensusMessage, ConsensusMsgPayload
from bxgateway.utils.block_header_info import BlockHeaderInfo
from bxgateway.utils.block_info import BlockInfo
from bxgateway.utils.errors import message_conversion_error

logger = logging.get_logger(__name__)


def parse_bx_block_header(bx_block: memoryview, block_pieces: Deque[Union[bytearray, memoryview]]) -> \
        BlockHeaderInfo:
    block_offsets = compact_block_short_ids_serializer.get_bx_block_offsets(bx_block)
    short_ids, short_ids_len = compact_block_short_ids_serializer.deserialize_short_ids_from_buffer(
        bx_block,
        block_offsets.short_id_offset
    )

    block_hash = OntObjectHash(binary=bx_block[block_offsets.block_begin_offset + 1:
                                               block_offsets.block_begin_offset + ont_constants.ONT_HASH_LEN + 1])
    offset = block_offsets.block_begin_offset + ont_constants.ONT_HASH_LEN + 1
    txn_count, = struct.unpack_from("<L", bx_block, offset)
    offset += ont_constants.ONT_INT_LEN
    payload_tail_len, = struct.unpack_from("<L", bx_block, offset)
    offset += ont_constants.ONT_INT_LEN + payload_tail_len
    owner_and_signature_len, = struct.unpack_from("<L", bx_block, offset)
    offset += ont_constants.ONT_INT_LEN + owner_and_signature_len
    consensus_payload_header_len, = struct.unpack_from("<L", bx_block, offset)
    offset += ont_constants.ONT_INT_LEN
    block_pieces.append(bx_block[offset: offset + consensus_payload_header_len])
    offset += consensus_payload_header_len
    block_pieces.append(bx_block[offset: offset + ont_constants.ONT_CHAR_LEN])
    offset += ont_constants.ONT_CHAR_LEN
    block_pieces.append(bx_block[offset: offset + ont_constants.ONT_INT_LEN])
    offset += ont_constants.ONT_INT_LEN
    block_start_len_and_txn_header_total_len, = struct.unpack_from("<L", bx_block, offset)
    offset += ont_constants.ONT_INT_LEN
    block_pieces.append(bx_block[offset: offset + block_start_len_and_txn_header_total_len])
    offset += block_start_len_and_txn_header_total_len

    return BlockHeaderInfo(block_offsets, short_ids, short_ids_len, block_hash, offset, txn_count)


def parse_bx_block_transactions_and_msg_tail(block_hash: Sha256Hash, bx_block: memoryview, offset: int,
                                             short_ids: List[int], block_offsets: BlockOffsets,
                                             tx_service: TransactionService,
                                             block_pieces: Deque[Union[bytearray, memoryview]]) -> \
        Tuple[List[int], List[Sha256Hash], int]:
    has_missing, unknown_tx_sids, unknown_tx_hashes = tx_service.get_missing_transactions(short_ids)
    if has_missing:
        return unknown_tx_sids, unknown_tx_hashes, offset
    short_tx_index = 0
    output_offset = offset
    while offset < block_offsets.short_id_offset:
        if bx_block[offset] == ont_constants.ONT_SHORT_ID_INDICATOR:
            try:
                sid = short_ids[short_tx_index]
            except IndexError:
                raise message_conversion_error.btc_block_decompression_error(
                    block_hash,
                    f"Message is improperly formatted, short id index ({short_tx_index}) "
                    f"exceeded its array bounds (size: {len(short_ids)})"
                )
            tx_hash, tx, _ = tx_service.get_transaction(sid)
            offset += ont_constants.ONT_SHORT_ID_INDICATOR_LENGTH
            short_tx_index += 1
        else:
            tx_size = ont_messages_util.get_next_tx_size(bx_block, offset)
            tx = bx_block[offset:offset + tx_size]
            offset += tx_size

        assert tx is not None
        block_pieces.append(tx)
        output_offset += len(tx)

    # Add consensus payload tail and owner and signature to block_pieces
    offset = block_offsets.block_begin_offset + ont_constants.ONT_HASH_LEN + ont_constants.ONT_INT_LEN + 1
    payload_tail_len, = struct.unpack_from("<L", bx_block, offset)
    offset += ont_constants.ONT_INT_LEN
    block_pieces.append(bx_block[offset: offset + payload_tail_len])
    offset += payload_tail_len
    owner_and_signature_len, = struct.unpack_from("<L", bx_block, offset)
    offset += ont_constants.ONT_INT_LEN
    block_pieces.append(bx_block[offset: offset + owner_and_signature_len])
    offset += owner_and_signature_len

    return unknown_tx_sids, unknown_tx_hashes, output_offset


def build_ont_block(block_pieces: Deque[Union[bytearray, memoryview]]) -> OntConsensusMessage:
    ont_block = bytearray()
    ont_block += block_pieces[0]
    # Construct base 64 encoded data and then add it to ont_block
    consensus_data_type, = struct.unpack_from("<B", block_pieces[1], 0)
    consensus_data_len, = struct.unpack_from("<L", block_pieces[2], 0)
    consensus_data_payload = bytearray(block_pieces[3])
    for i in range(4, len(block_pieces) - 1):
        consensus_data_payload += block_pieces[i]
    consensus_data = ConsensusMsgPayload(consensus_data_type, consensus_data_len,
                                         base64.b64encode(
                                             bytes(consensus_data_payload)
                                         ).decode(constants.DEFAULT_TEXT_ENCODING))
    consensus_msg_payload = json.dumps(consensus_data, cls=EnhancedJSONEncoder, separators=(",", ":"))
    ont_block += bytearray(consensus_msg_payload.encode(constants.DEFAULT_TEXT_ENCODING))
    ont_block += block_pieces[len(block_pieces) - 1]
    return OntConsensusMessage(buf=ont_block)


class OntNormalConsensusMessageConverter(AbstractOntMessageConverter):

    def block_to_bx_block(
        self,
        block_msg: OntConsensusMessage,
        tx_service: TransactionService,
        enable_block_compression: bool,
        min_tx_age_seconds: float
    ) -> Tuple[memoryview, BlockInfo]:
        """
        Pack an Ontology consensus message's transactions into a bloXroute block.
        """
        consensus_msg = block_msg
        compress_start_datetime = datetime.utcnow()
        compress_start_timestamp = time.time()
        size = 0
        buf = deque()
        short_ids = []
        ignored_sids = []
        original_size = len(consensus_msg.rawbytes())

        consensus_payload_header = consensus_msg.consensus_payload_header()
        consensus_payload_header_len = bytearray(ont_constants.ONT_INT_LEN)
        struct.pack_into("<L", consensus_payload_header_len, 0, len(consensus_payload_header))
        size += ont_constants.ONT_INT_LEN
        buf.append(consensus_payload_header_len)
        size += len(consensus_payload_header)
        buf.append(consensus_payload_header)
        consensus_data_type = bytearray(ont_constants.ONT_CHAR_LEN)
        struct.pack_into("<B", consensus_data_type, 0, consensus_msg.consensus_data_type())
        size += ont_constants.ONT_CHAR_LEN
        buf.append(consensus_data_type)
        consensus_data_len = bytearray(ont_constants.ONT_INT_LEN)
        struct.pack_into("<L", consensus_data_len, 0, consensus_msg.consensus_data_len())
        size += ont_constants.ONT_INT_LEN
        buf.append(consensus_data_len)
        block_start_len = consensus_msg.block_start_len_memoryview()
        txn_header = consensus_msg.txn_header()
        block_start_len_and_txn_header_total_len = bytearray(ont_constants.ONT_INT_LEN)
        struct.pack_into("<L", block_start_len_and_txn_header_total_len, 0, len(block_start_len) + len(txn_header))
        size += ont_constants.ONT_INT_LEN
        buf.append(block_start_len_and_txn_header_total_len)
        size += len(block_start_len)
        buf.append(block_start_len)
        size += len(txn_header)
        buf.append(txn_header)
        max_timestamp_for_compression = time.time() - min_tx_age_seconds

        for tx in consensus_msg.txns():
            tx_hash, _ = ont_messages_util.get_txid(tx)
            transaction_key = tx_service.get_transaction_key(tx_hash)
            short_id = tx_service.get_short_id_by_key(transaction_key)
            short_id_assign_time = 0

            if short_id != constants.NULL_TX_SID:
                short_id_assign_time = tx_service.get_short_id_assign_time(short_id)

            if short_id == constants.NULL_TX_SID or \
                    not enable_block_compression or \
                    short_id_assign_time > max_timestamp_for_compression:
                if short_id != constants.NULL_TX_SIDS:
                    ignored_sids.append(short_id)
                buf.append(tx)
                size += len(tx)
            else:
                short_ids.append(short_id)
                buf.append(ont_constants.ONT_SHORT_ID_INDICATOR_AS_BYTEARRAY)
                size += 1

        # Prepend owner and signature, consensus payload tail, tx count and block hash to bx_block
        owner_and_signature = consensus_msg.owner_and_signature()
        owner_and_signature_len = bytearray(ont_constants.ONT_INT_LEN)
        struct.pack_into("<L", owner_and_signature_len, 0, len(owner_and_signature))
        size += len(owner_and_signature)
        buf.appendleft(owner_and_signature)
        size += ont_constants.ONT_INT_LEN
        buf.appendleft(owner_and_signature_len)
        payload_tail = consensus_msg.payload_tail()
        payload_tail_len = bytearray(ont_constants.ONT_INT_LEN)
        struct.pack_into("<L", payload_tail_len, 0, len(payload_tail))
        size += len(payload_tail)
        buf.appendleft(payload_tail)
        size += ont_constants.ONT_INT_LEN
        buf.appendleft(payload_tail_len)
        txn_count = bytearray(ont_constants.ONT_INT_LEN)
        struct.pack_into("<L", txn_count, 0, consensus_msg.txn_count())
        size += ont_constants.ONT_INT_LEN
        buf.appendleft(txn_count)
        block_hash = consensus_msg.block_hash().binary
        size += ont_constants.ONT_HASH_LEN
        buf.appendleft(block_hash)

        is_consensus_msg_buf = struct.pack("?", True)
        buf.appendleft(is_consensus_msg_buf)
        size += 1

        block = finalize_block_bytes(buf, size, short_ids)

        prev_block_hash = convert.bytes_to_hex(consensus_msg.prev_block_hash().binary)
        bx_block_hash = convert.bytes_to_hex(crypto.double_sha256(block))

        block_info = BlockInfo(
            consensus_msg.block_hash(),
            short_ids,
            compress_start_datetime,
            datetime.utcnow(),
            (time.time() - compress_start_timestamp) * 1000,
            consensus_msg.txn_count(),
            bx_block_hash,
            prev_block_hash,
            original_size,
            size,
            100 - float(size) / original_size * 100,
            ignored_sids
        )

        return memoryview(block), block_info

    def bx_block_to_block(self, bx_block_msg: memoryview, tx_service: TransactionService) -> BlockDecompressionResult:
        """
        Uncompresses a bx_block from a broadcast bx_block message and converts to a raw Ontology bx_block.

        bx_block must be a memoryview, since memoryview[offset] returns a bytearray, while bytearray[offset] returns
        a byte.
        """
        decompress_start_datetime = datetime.utcnow()
        decompress_start_timestamp = time.time()

        # Initialize tracking of transaction and SID mapping
        block_pieces = deque()
        header_info = parse_bx_block_header(bx_block_msg, block_pieces)
        unknown_tx_sids, unknown_tx_hashes, offset = parse_bx_block_transactions_and_msg_tail(
            header_info.block_hash,
            bx_block_msg,
            header_info.offset,
            header_info.short_ids,
            header_info.block_offsets,
            tx_service,
            block_pieces
        )
        total_tx_count = header_info.txn_count

        if not unknown_tx_sids and not unknown_tx_hashes:
            ont_consensus_msg = build_ont_block(block_pieces)
            logger.debug("Successfully parsed bx_block broadcast message. {} transactions in bx_block", total_tx_count)
        else:
            ont_consensus_msg = None
            logger.warning(log_messages.BLOCK_RECOVERY_NEEDED_ONT_CONSENSUS,
                           len(unknown_tx_sids), len(unknown_tx_hashes), total_tx_count)
        block_info = get_block_info(
            bx_block_msg,
            header_info.block_hash,
            header_info.short_ids,
            decompress_start_datetime,
            decompress_start_timestamp,
            total_tx_count,
            ont_consensus_msg
        )
        return BlockDecompressionResult(ont_consensus_msg, block_info, unknown_tx_sids, unknown_tx_hashes)



import time
from heapq import heappop, heappush
from threading import RLock
from typing import List, Optional, Callable, Dict

from bxcommon import constants
from bxcommon.utils import performance_utils
from bxutils import logging
from bxutils.logging import LogRecordType

logger = logging.get_logger(__name__)
alarm_troubleshooting_logger = logging.get_logger(LogRecordType.AlarmTroubleshooting, __name__)


class Alarm:
    """
    Alarm object. Encapsulates function and arguments.
    """

    def __init__(
        self, fn: Callable, fire_time: float, *args, name: Optional[str] = None, **kwargs
    ):
        if name is None:
            if hasattr(fn, "im_class"):
                cls = getattr(fn, "im_class")
                name = f"{cls}#{fn.__name__}"

            else:
                name = fn.__name__

        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.fire_time = fire_time
        self.name = name

    def fire(self):
        time_from_expected = time.time() - self.fire_time
        if time_from_expected > constants.WARN_ALARM_EXECUTION_OFFSET:
            logger.debug("{} executed {} seconds later than expected.", self, time_from_expected)
        return self.fn(*self.args, **self.kwargs)

    def __repr__(self):
        return f"Alarm<function: {self.name}, fire_time: {self.fire_time}"


class AlarmId:
    fire_time: float
    count: int
    alarm: Alarm
    is_active: bool

    def __init__(self, fire_time: float, count: int, alarm: Alarm) -> None:
        self.fire_time = fire_time
        self.count = count
        self.alarm = alarm
        self.is_active = True

    def __repr__(self):
        return (
            f"AlarmId<"
            f"fire_time: {self.fire_time}, "
            f"count: {self.count}, "
            f"active: {self.is_active}, "
            f"alarm: {self.alarm}"
            f">"
        )

    def __lt__(self, other):
        if not isinstance(other, AlarmId):
            raise TypeError("< not supported between instances of 'AlarmId' and {}".format(other))

        return self.fire_time < other.fire_time or (self.fire_time == other.fire_time and self.count < other.count)

    def __eq__(self, other):
        if not isinstance(other, AlarmId):
            return False
        return (
            self.fire_time == other.fire_time
            and self.count == other.count
            and self.alarm == other.alarm
            and self.is_active == other.is_active
        )


class AlarmQueue:
    """
    Queue for events that take place some time in the future.

    Constants
    ---------
    REMOVED: canceled alarm id for remove

    Attributes
    ----------
    alarms: list of alarms, denoted by (fire time, unique count, alarm function)
    uniq_count: counter used for tiebreakers in heap comparison if same fire time
    approx_alarms_scheduled: function => min-heap of scheduled alarms. used to ensure multiple alarms
                             with the same function handle are not executed repeatedly
    """

    def __init__(self) -> None:
        self.alarms: List[AlarmId] = []
        self.uniq_count: int = 0
        self.approx_alarms_scheduled: Dict[Callable, List[AlarmId]] = {}
        self.lock = RLock()

    def register_alarm(
        self,
        fire_delay: float,
        fn: Callable,
        *args,
        alarm_name: Optional[str] = None,
        fire_immediately: bool = False,
        **kwargs
    ) -> AlarmId:
        """
        Schedules an alarm to be fired in `fire_delay` seconds. Function must return a positive integer
        to be schedule again in the future.
        :param fire_delay: delay in seconds before firing alarm
        :param fn: function to be fired on delay
        :param alarm_name: optional label for alarm
        :param fire_immediately: fires alarm immediately if delay is 0
        :param args: function arguments
        :return: (fire time, unique count, alarm function)
        """
        try:
            hash(fn)
        except Exception:
            raise ValueError(f"Could not register an unhashable alarm: {fn}")

        if fire_delay < 0:
            raise ValueError("Invalid negative fire delay.")
        if fn is None:
            raise ValueError("Function cannot be None.")

        alarm = Alarm(fn, time.time() + fire_delay, *args, name=alarm_name, **kwargs)
        alarm_id = AlarmId(time.time() + fire_delay, self.uniq_count, alarm)

        if fire_immediately and fire_delay == 0:
            alarm.fire()
            return alarm_id

        with self.lock:
            heappush(self.alarms, alarm_id)
            self.uniq_count += 1
        return alarm_id

    def register_approx_alarm(
        self,
        fire_delay: float,
        slop: float,
        fn: Callable,
        *args,
        alarm_name: Optional[str] = None,
        **kwargs
    ) -> None:
        """
        Schedules an alarm that will fire in between `fire_delay` +/- slop seconds. The referenced function
        will only be executed once in that time interval, even if multiple calls to this method are made.
        :param fire_delay: delay in seconds before firing alarm
        :param slop: range during which only once instance of this function may be fire
        :param fn: function to be fired
        :param args: function arguments
        :param alarm_name: name of alarm for logging
        """
        try:
            hash(fn)
        except Exception:
            raise ValueError(f"Could not register an unhashable alarm: {fn}")

        if fire_delay < 0:
            raise ValueError("Invalid negative fire delay.")
        if slop < 0:
            raise ValueError("Invalid negative slop.")

        with self.lock:
            if fn not in self.approx_alarms_scheduled:
                new_alarm_id = self.register_alarm(
                    fire_delay, fn, *args, alarm_name=alarm_name, **kwargs
                )
                self.approx_alarms_scheduled[fn] = []
                heappush(self.approx_alarms_scheduled[fn], new_alarm_id)
            else:
                now = time.time()
                late_time, early_time = fire_delay + now + slop, fire_delay + now - slop
                for alarm_id in self.approx_alarms_scheduled[fn]:
                    if early_time <= alarm_id.fire_time <= late_time:
                        return
                heappush(
                    self.approx_alarms_scheduled[fn],
                    self.register_alarm(fire_delay, fn, *args, alarm_name=alarm_name, **kwargs)
                )

    def unregister_alarm(self, alarm_id: AlarmId) -> None:
        """
        Cancels alarm and cleans up the head of the heap.
        :param alarm_id: alarm to cancel
        """
        alarm_id.is_active = False

        with self.lock:
            while self.alarms and not self.alarms[0].is_active:
                heappop(self.alarms)

    def fire_alarms(self) -> None:
        """
        Fire alarms that are ready.
        Reschedules alarms that return a value > 0 with the return value as the next timeout.
        """
        if not self.alarms:
            return

        curr_time = time.time()
        alarms_count = 0

        with self.lock:
            while self.alarms and self.alarms[0].fire_time <= curr_time:
                alarm_id: AlarmId = heappop(self.alarms)
                alarm = alarm_id.alarm

                if alarm_id.is_active:
                    try:
                        start_time = time.time()
                        next_delay = alarm.fire()
                        alarms_count += 1
                    # pylint: disable=broad-except
                    except Exception as e:
                        logger.exception("Alarm {} could not fire and failed with exception: {}", alarm, e)
                        if alarm.fn in self.approx_alarms_scheduled:
                            self._pop_and_cleanup_approx_alarm(alarm_id)
                    else:
                        performance_utils.log_operation_duration(
                            alarm_troubleshooting_logger,
                            "Single alarm", start_time,
                            constants.WARN_ALARM_EXECUTION_DURATION,
                            alarm=alarm
                        )

                        if next_delay is not None and next_delay > 0:
                            next_time = time.time() + next_delay
                            alarm_id.fire_time = next_time
                            alarm_id.alarm.fire_time = next_time
                            heappush(self.alarms, alarm_id)
                        # Delete alarm from approx_alarms_scheduled if applicable
                        elif alarm.fn in self.approx_alarms_scheduled:
                            self._pop_and_cleanup_approx_alarm(alarm_id)

        performance_utils.log_operation_duration(
            alarm_troubleshooting_logger,
            "Alarms",
            curr_time,
            constants.WARN_ALL_ALARMS_EXECUTION_DURATION,
            count=alarms_count
        )

    def fire_ready_alarms(self) -> Optional[float]:
        """
        Fires ready alarm repeatedly until no more should be fired.
        :return: time until next alarm, or None
        """
        time_to_next_alarm = self.time_to_next_alarm()

        while time_to_next_alarm is not None and time_to_next_alarm <= 0:
            self.fire_alarms()
            time_to_next_alarm = self.time_to_next_alarm()

        return time_to_next_alarm

    def time_to_next_alarm(self) -> Optional[float]:
        """
        Indicates if there's not an alarm on the queue and the timeout to the next one if there is.
        :return: (if alarm queue is empty, timeout to next alarm)
        """
        with self.lock:
            if not self.alarms:
                return None

            time_to_alarm = self.alarms[0].fire_time - time.time()
            return time_to_alarm

    def _pop_and_cleanup_approx_alarm(self, alarm_id: AlarmId) -> None:
        alarm_heap = self.approx_alarms_scheduled[alarm_id.alarm.fn]

        # alarm that was just fired must be the first one in the list
        assert alarm_heap[0].count == alarm_id.count
        heappop(alarm_heap)

        if not alarm_heap:
            del self.approx_alarms_scheduled[alarm_id.alarm.fn]



import argparse
import sys
from typing import Tuple

from bxcommon.models.blockchain_peer_info import BlockchainPeerInfo
from bxcommon.models.blockchain_protocol import BlockchainProtocol
from bxcommon.utils.blockchain_utils.eth import eth_common_constants
from bxgateway import log_messages
from bxutils import logging


logger = logging.get_logger(__name__)


# Make sure enode is at least as long as the public key
def enode_is_valid_length(enode: str) -> bool:
    return len(enode) >= 2 * eth_common_constants.PUBLIC_KEY_LEN


def get_enode_parts(enode: str) -> Tuple[str, str, str]:
    enode_and_pub_key, ip_and_port = enode.split("@")
    if enode_and_pub_key.startswith("enode://"):
        pub_key = enode_and_pub_key[8:]
    else:
        pub_key = enode_and_pub_key
    ip, port_and_disc = ip_and_port.split(":")
    port = port_and_disc.split("?")[0]
    return pub_key, ip, port


def get_ip_port_string_parts(ip_port_string: str) -> Tuple[str, str]:
    ip_port_list = ip_port_string.strip().split(":")
    ip = ip_port_list[0]
    port = ip_port_list[1]
    return ip, port


def parse_enode(enode: str) -> BlockchainPeerInfo:
    if not enode_is_valid_length(enode):
        logger.fatal(log_messages.ETH_PARSER_INVALID_ENODE_LENGTH, enode, len(enode), exc_info=False)
        sys.exit(1)
    try:
        pub_key, ip, port = get_enode_parts(enode)
        if not port.isnumeric():
            logger.fatal(log_messages.PARSER_INVALID_PORT, port, exc_info=False)
            sys.exit(1)
    except ValueError:
        logger.fatal(log_messages.ETH_PARSER_INVALID_ENODE, enode, exc_info=False)
        sys.exit(1)
    else:
        return BlockchainPeerInfo(ip, int(port), pub_key)


def parse_ip_port(ip_port_string: str) -> BlockchainPeerInfo:
    ip, port = get_ip_port_string_parts(ip_port_string)
    if not port.isnumeric():
        logger.fatal(log_messages.PARSER_INVALID_PORT, port, exc_info=False)
        sys.exit(1)
    return BlockchainPeerInfo(ip, int(port))


def parse_peer(blockchain_protocol: str, peer: str) -> BlockchainPeerInfo:
    if blockchain_protocol.lower() == BlockchainProtocol.ETHEREUM.value:
        return parse_enode(peer)
    else:
        return parse_ip_port(peer)


class ParseEnode(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        assert isinstance(values, str)
        blockchain_peer = parse_enode(values)
        # Node public key gets validated in validate_eth_opts
        namespace.node_public_key = blockchain_peer.node_public_key
        # blockchain IP gets validated in __init__()
        namespace.blockchain_ip = blockchain_peer.ip
        namespace.blockchain_port = blockchain_peer.port


class ParseBlockchainPeers(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        assert isinstance(values, str)
        blockchain_peers = set()
        blockchain_protocol = namespace.blockchain_protocol
        for peer in values.split(","):
            blockchain_peer = parse_peer(blockchain_protocol, peer)
            blockchain_peers.add(blockchain_peer)
        namespace.blockchain_peers = blockchain_peers



import struct

from bxcommon.utils import convert
from bxcommon.utils.blockchain_utils.btc import btc_common_utils
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash
from bxgateway.btc_constants import BTC_HDR_COMMON_OFF
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.btc import btc_messages_util


def pack_outpoint(hash_val, index, buf, off):
    return struct.pack_into("<32sI", buf, off, hash_val, index)


# A transaction input.
# This class cannot parse a transaction input from rawbytes, but can construct one.
class TxIn(object):
    def __init__(self, prev_outpoint_hash=None, prev_out_index=None, sig_script=None, sequence=None, buf=None, off=None,
                 length=None):
        if buf is None:
            buf = bytearray(36 + 9 + len(sig_script) + 4)
            self.buf = buf

            off = 0
            pack_outpoint(prev_outpoint_hash, prev_out_index, buf, off)
            off += 36
            off += btc_messages_util.pack_int_to_btc_varint(len(sig_script), buf, off)
            buf[off:off + len(sig_script)] = sig_script
            off += len(sig_script)
            struct.pack_into("<I", buf, off, sequence)
            off += 4
            self.size = off
            self.off = 0
        else:
            self.buf = buf
            self.size = length
            self.off = off

        self._memoryview = memoryview(buf)

    def rawbytes(self):
        if self.size == len(self.buf) and self.off == 0:
            return self.buf
        else:
            return self._memoryview[self.off:self.off + self.size]


# A transaction output.
# This class cannot parse a transaction output from rawbytes, but can construct one.
class TxOut(object):
    def __init__(self, value=None, pk_script=None, buf=None, off=None, length=None):
        if buf is None:
            pk_script_len = len(pk_script)
            buf = bytearray(8 + 9 + pk_script_len)
            self.buf = buf

            off = 0
            struct.pack_into("<Q", buf, off, value)
            off += 8
            off += btc_messages_util.pack_int_to_btc_varint(pk_script_len, buf, off)
            buf[off:off + pk_script_len]= pk_script
            self.size = off + pk_script_len
            self.off = 0
        else:
            self.buf = buf
            self.size = length
            self.off = off

        self._memoryview = memoryview(buf)

    def rawbytes(self):
        if self.size == len(self.buf) and self.off == 0:
            return self.buf
        else:
            return self._memoryview[self.off:self.off + self.size]


class TxBtcMessage(BtcMessage):
    """
    Transaction message. This class cannot fully parse a transaction message from bytes, but can construct one.

    Attributes
    ----------
    tx_in: a list of TxIn instances
    tx_out: a list of TxOut instances
    """
    MESSAGE_TYPE = BtcMessageType.TRANSACTIONS

    def __init__(self, magic=None, version=None, tx_in=None, tx_out=None, lock_time=None, is_segwit=None, buf=None):
        if buf is None:
            tot_size = sum([len(x.rawbytes()) for x in tx_in]) + sum([len(x.rawbytes()) for x in tx_out])
            buf = bytearray(BTC_HDR_COMMON_OFF + 2 * 9 + 8 + tot_size)
            self.buf = buf

            off = BTC_HDR_COMMON_OFF
            struct.pack_into("<I", buf, off, version)
            off += 4
            off += btc_messages_util.pack_int_to_btc_varint(len(tx_in), buf, off)

            for inp in tx_in:
                rawbytes = inp.rawbytes()
                size = len(rawbytes)
                buf[off:off + size] = rawbytes
                off += size

            off += btc_messages_util.pack_int_to_btc_varint(len(tx_out), buf, off)

            for out in tx_out:
                rawbytes = out.rawbytes()
                size = len(rawbytes)
                buf[off:off + size] = rawbytes
                off += size

            struct.pack_into("<I", buf, off, lock_time)
            off += 4

            super(TxBtcMessage, self).__init__(magic, self.MESSAGE_TYPE, off - BTC_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._version = None
        self._tx_in = None
        self._tx_out = None
        self._lock_time = None
        self._tx_hash = None
        self._tx_id = None
        self._tx_out_count = None
        self._tx_in_count = None
        self._is_segwit_tx = is_segwit

    def version(self):
        if self._version is None:
            off = BTC_HDR_COMMON_OFF
            self._version, = struct.unpack_from('<i', self.buf, off)
        return self._version

    def tx_in(self):
        if self._tx_in is None:
            off = BTC_HDR_COMMON_OFF + 4
            self._tx_in_count, size = btc_common_utils.btc_varint_to_int(self.buf, off)
            off += size
            self._tx_in = []

            start = off
            end = off

            for _ in range(self._tx_in_count):
                end += 36
                script_len, size = btc_common_utils.btc_varint_to_int(self.buf, end)
                end += size + script_len + 4
                self._tx_in.append(self.rawbytes()[start:end])
                start = end

            off = end
            self._tx_out_count, size = btc_common_utils.btc_varint_to_int(self.buf, off)
            self._tx_out = []
            off += size

            start = off
            end = off
            for _ in range(self._tx_out_count):
                end += 8
                script_len, size = btc_common_utils.btc_varint_to_int(self.buf, end)
                end += size + script_len
                self._tx_out.append(self.rawbytes()[start:end])

            off = end

            self._lock_time, = struct.unpack_from('<I', self.buf, off)

        return self._tx_in

    def tx_out(self):
        if self._tx_in is None:
            self.tx_in()
        return self._tx_out

    def lock_time(self):
        if self._tx_in is None:
            self.tx_in()
        return self._lock_time

    def is_segwit_tx(self) -> bool:
        """
        Determines if a transaction is a segwit transaction by reading the marker and flag bytes
        :return: boolean indicating segwit
        """
        if self._is_segwit_tx is None:
            self._is_segwit_tx = btc_common_utils.is_segwit(self.payload())
        return self._is_segwit_tx

    def tx_hash(self) -> BtcObjectHash:
        """
        Actually gets the txid, which is the same as the hash for non segwit transactions
        :return: BtcObjectHash
        """
        if self._tx_hash is None:
            self._tx_hash = btc_common_utils.get_txid(self.payload())
        # pyre-fixme[7]: Expected `BtcObjectHash` but got `None`.
        return self._tx_hash

    def tx(self):
        return self.payload()

    def __repr__(self):
        return ("TxBtcMessage<version: {}, length: {}, tx_hash: {}, rawbytes: {}>"
                .format(self.version(),
                        len(self.rawbytes()),
                        convert.bytes_to_hex(self.tx_hash().binary),
                        convert.bytes_to_hex(
                            self.rawbytes().tobytes())))



import weakref
from typing import TYPE_CHECKING

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.connections.ont.ont_remote_connection_protocol import OntRemoteConnectionProtocol
from bxgateway.messages.ont.ont_message_factory import ont_message_factory
from bxgateway.messages.ont.ping_ont_message import PingOntMessage

if TYPE_CHECKING:
    from bxgateway.connections.ont.ont_gateway_node import OntGatewayNode


class OntRemoteConnection(AbstractGatewayBlockchainConnection["OntGatewayNode"]):
    CONNECTION_TYPE = ConnectionType.REMOTE_BLOCKCHAIN_NODE

    def __init__(self, socket_connection: AbstractSocketConnectionProtocol, node: "OntGatewayNode"):
        super(OntRemoteConnection, self).__init__(socket_connection, node)
        self.connection_protocol = weakref.ref(OntRemoteConnectionProtocol(self))

    def connection_message_factory(self) -> AbstractMessageFactory:
        return ont_message_factory

    def ping_message(self) -> AbstractMessage:
        return PingOntMessage(
            self.node.opts.blockchain_net_magic, height=self.node.current_block_height
        )



from typing import Union, cast, List, NamedTuple, Set, Optional, TYPE_CHECKING

from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.messages.bloxroute.txs_message import TxsMessage
from bxcommon.models.blockchain_protocol import BlockchainProtocol
from bxcommon.models.transaction_key import TransactionKey
from bxcommon.models.tx_validation_status import TxValidationStatus
from bxcommon.services.transaction_service import TransactionService, TransactionCacheKeyType, wrap_sha256, \
    TransactionFromBdnGatewayProcessingResult
from bxcommon.utils.blockchain_utils import transaction_validation
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon import constants
from bxgateway.abstract_message_converter import AbstractMessageConverter
from bxgateway.messages.btc.tx_btc_message import TxBtcMessage
from bxgateway.messages.eth.protocol.transactions_eth_protocol_message import TransactionsEthProtocolMessage
from bxgateway.services.block_recovery_service import RecoveredTxsSource

if TYPE_CHECKING:
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


class ProcessTransactionMessageFromNodeResult(NamedTuple):
    seen: bool
    transaction_hash: Sha256Hash
    transaction_contents: Union[bytearray, memoryview]
    bdn_transaction_message: TxMessage
    tx_validation_status: TxValidationStatus


class OntTxMessage(object):
    pass


class MissingTransactions(NamedTuple):
    short_id: int
    transaction_hash: Sha256Hash


class GatewayTransactionService(TransactionService):

    node: "AbstractGatewayNode"

    def process_gateway_transaction_from_bdn(
        self,
        transaction_hash: Sha256Hash,
        short_id: int,
        transaction_contents: Union[bytearray, memoryview],
        is_compact: bool
    ) -> TransactionFromBdnGatewayProcessingResult:

        transaction_key = self.get_transaction_key(transaction_hash)
        if (
            (not short_id or (short_id and self.has_short_id(short_id)))
            and self.has_transaction_contents_by_key(transaction_key)
            or self.removed_transaction_by_key(transaction_key)
        ):
            return TransactionFromBdnGatewayProcessingResult(ignore_seen=True)

        existing_short_ids = self.get_short_ids_by_key(transaction_key)
        if (
            (self.has_transaction_contents_by_key(transaction_key) or is_compact)
            and short_id and short_id in existing_short_ids
        ):
            return TransactionFromBdnGatewayProcessingResult(existing_short_id=True)

        assigned_short_id = False
        set_content = False

        if short_id and short_id not in existing_short_ids:
            self.assign_short_id_by_key(transaction_key, short_id)
            assigned_short_id = True

        existing_contents = self.has_transaction_contents_by_key(transaction_key)

        if not is_compact and not existing_contents:
            self.set_transaction_contents_by_key(transaction_key, transaction_contents)
            set_content = True

        return TransactionFromBdnGatewayProcessingResult(
            assigned_short_id=assigned_short_id,
            existing_contents=existing_contents,
            set_content=set_content
        )

    def process_transactions_message_from_node(
        self,
        msg: Union[TxBtcMessage, TransactionsEthProtocolMessage, OntTxMessage],
        min_tx_network_fee: int,
        enable_transaction_validation: bool
    ) -> List[ProcessTransactionMessageFromNodeResult]:
        message_converter = cast(AbstractMessageConverter, self.node.message_converter)

        account_id = self.node.account_id if self.node.account_id is not None else constants.DECODED_EMPTY_ACCOUNT_ID
        # avoid filtering low fee transactions in tx_to_bx_txs to match extensions behavior
        bx_tx_messages = message_converter.tx_to_bx_txs(
            msg, self.network_num, self.node.default_tx_flag, min_tx_network_fee=0, account_id=account_id
        )

        result = []
        blockchain_protocol = BlockchainProtocol(self.network.protocol.lower())

        for bx_tx_message, tx_hash, tx_bytes in bx_tx_messages:
            transaction_key = self.get_transaction_key(tx_hash)
            tx_cache_key = self._tx_hash_to_cache_key(tx_hash)
            tx_seen_flag = (
                self.has_transaction_contents_by_cache_key(tx_cache_key)
                or self.removed_transaction_by_cache_key(tx_cache_key)
            )

            if not tx_seen_flag:
                self.set_transaction_contents_by_key(transaction_key, tx_bytes)

            tx_validation_status = TxValidationStatus.VALID_TX
            if enable_transaction_validation:
                tx_validation_status = transaction_validation.validate_transaction(
                    tx_bytes, blockchain_protocol, min_tx_network_fee
                )

            result.append(
                ProcessTransactionMessageFromNodeResult(
                    tx_seen_flag, tx_hash, tx_bytes, bx_tx_message, tx_validation_status
                )
            )

        return result

    def process_txs_message(
        self,
        msg: TxsMessage
    ) -> Set[MissingTransactions]:
        missing_transactions: Set[MissingTransactions] = set()
        transactions = msg.get_txs()
        missing: bool
        for transaction in transactions:
            transaction_hash, transaction_contents, short_id = transaction
            missing = False

            assert transaction_hash is not None
            assert short_id is not None

            transaction_key = self.get_transaction_key(transaction_hash)

            if not self.has_short_id(short_id):
                self.assign_short_id_by_key(transaction_key, short_id)
                missing = True

            if not self.has_transaction_contents_by_key(transaction_key):
                assert transaction_contents is not None
                self.set_transaction_contents_by_key(transaction_key, transaction_contents)

                missing = True

            if missing:
                missing_transactions.add(MissingTransactions(short_id, transaction_hash))

        return missing_transactions

    def set_transaction_contents_base_by_key(
        self,
        transaction_key: TransactionKey,
        has_short_id: bool,
        previous_size: int,
        call_set_contents: bool,
        transaction_contents: Optional[Union[bytearray, memoryview]] = None,
        transaction_contents_length: Optional[int] = None
    ) -> None:
        """
        Adds transaction contents to transaction service cache with lookup key by transaction hash

        :param transaction_key: transaction key
        :param has_short_id: flag indicating if cache already has short id for given transaction
        :param previous_size: previous size of transaction contents if already exists
        :param call_set_contents: flag indicating if method should make a call to set content form Python code
        :param transaction_contents: transaction contents bytes
        :param transaction_contents_length: if the transaction contents bytes not available, just send the length
        """
        super(GatewayTransactionService, self).set_transaction_contents_base_by_key(
            transaction_key,
            has_short_id,
            previous_size,
            call_set_contents,
            transaction_contents,
            transaction_contents_length
        )
        if transaction_contents is not None:
            self.node.log_txs_network_content(self.network_num, transaction_key.transaction_hash, transaction_contents)
            if call_set_contents:
                self.node.block_recovery_service.check_missing_tx_hash(
                    transaction_key.transaction_hash, RecoveredTxsSource.TXS_RECEIVED_FROM_NODE
                )



import time
from abc import ABCMeta
from asyncio import Future
from typing import TYPE_CHECKING, Set, Optional

from bxcommon import constants
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.connections.internal_node_connection import InternalNodeConnection
from bxcommon.feed.feed import FeedKey
from bxcommon.messages.bloxroute.abstract_cleanup_message import AbstractCleanupMessage
from bxcommon.messages.bloxroute.bdn_performance_stats_message import BdnPerformanceStatsMessage
from bxcommon.messages.bloxroute.blockchain_network_message import RefreshBlockchainNetworkMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.bloxroute_message_validator import BloxrouteMessageValidator
from bxcommon.messages.bloxroute.compressed_block_txs_message import CompressedBlockTxsMessage
from bxcommon.messages.bloxroute.disconnect_relay_peer_message import DisconnectRelayPeerMessage
from bxcommon.messages.bloxroute.hello_message import HelloMessage
from bxcommon.messages.bloxroute.notification_message import NotificationMessage
from bxcommon.messages.bloxroute.txs_message import TxsMessage
from bxcommon.messages.validation.message_size_validation_settings import MessageSizeValidationSettings
from bxcommon.models.entity_type_model import EntityType
from bxcommon.models.notification_code import NotificationCode
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.services import sdn_http_service
from bxcommon.utils import convert, performance_utils
from bxcommon.utils import memory_utils
from bxcommon.utils.alarm_queue import AlarmId
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats import hooks, stats_format
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxcommon.utils.stats.transaction_stat_event_type import TransactionStatEventType
from bxcommon.utils.stats.transaction_statistics_service import tx_stats
from bxgateway import log_messages, gateway_constants
from bxcommon.feed.new_transaction_feed import NewTransactionFeed, RawTransactionFeedEntry
from bxgateway.services.block_recovery_service import RecoveredTxsSource
from bxgateway.services.gateway_transaction_service import MissingTransactions
from bxgateway.utils.logging.status import status_log
from bxgateway.utils.stats.gateway_bdn_performance_stats_service import gateway_bdn_performance_stats_service, \
    GatewayBdnPerformanceStatInterval
from bxgateway.utils.stats.gateway_transaction_stats_service import gateway_transaction_stats_service
from bxgateway.utils.stats.transaction_feed_stats_service import transaction_feed_stats_service
from bxutils import logging
from bxutils.logging import LogLevel
from bxutils.logging import LogRecordType

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

msg_handling_logger = logging.get_logger(LogRecordType.MessageHandlingTroubleshooting, __name__)


class AbstractRelayConnection(InternalNodeConnection["AbstractGatewayNode"]):
    __metaclass__ = ABCMeta

    CONNECTION_TYPE = ConnectionType.RELAY_ALL

    node: "AbstractGatewayNode"

    def __init__(self, sock: AbstractSocketConnectionProtocol, node: "AbstractGatewayNode"):
        super(AbstractRelayConnection, self).__init__(sock, node)

        hello_msg = HelloMessage(protocol_version=self.protocol_version, network_num=self.network_num,
                                 node_id=self.node.opts.node_id)
        self.enqueue_msg(hello_msg)

        self.hello_messages = constants.BLOXROUTE_HELLO_MESSAGES
        self.header_size = constants.STARTING_SEQUENCE_BYTES_LEN + constants.BX_HDR_COMMON_OFF
        self.message_handlers = {
            BloxrouteMessageType.HELLO: self.msg_hello,
            BloxrouteMessageType.PING: self.msg_ping,
            BloxrouteMessageType.PONG: self.msg_pong,
            BloxrouteMessageType.ACK: self.msg_ack,
            BloxrouteMessageType.BROADCAST: self.msg_broadcast,
            BloxrouteMessageType.KEY: self.msg_key,
            BloxrouteMessageType.TRANSACTION: self.msg_tx,
            BloxrouteMessageType.TRANSACTIONS: self.msg_txs,
            BloxrouteMessageType.COMPRESSED_BLOCK_TXS: self.msg_compressed_block_txs,
            BloxrouteMessageType.BLOCK_HOLDING: self.msg_block_holding,
            BloxrouteMessageType.DISCONNECT_RELAY_PEER: self.msg_disconnect_relay_peer,
            BloxrouteMessageType.TX_SERVICE_SYNC_TXS: self.tx_sync_service.msg_tx_service_sync_txs,
            BloxrouteMessageType.TX_SERVICE_SYNC_COMPLETE: self.tx_sync_service.msg_tx_service_sync_complete,
            BloxrouteMessageType.BLOCK_CONFIRMATION: self.msg_cleanup,
            BloxrouteMessageType.TRANSACTION_CLEANUP: self.msg_cleanup,
            BloxrouteMessageType.NOTIFICATION: self.msg_notify,
            BloxrouteMessageType.REFRESH_BLOCKCHAIN_NETWORK: self.msg_refresh_blockchain_network
        }

        msg_size_validation_settings = MessageSizeValidationSettings(self.node.network.max_block_size_bytes,
                                                                     self.node.network.max_tx_size_bytes)
        self.message_validator = BloxrouteMessageValidator(msg_size_validation_settings, self.protocol_version)
        self.check_matching_relay_connection_alarm_id: Optional[AlarmId] = None

    def msg_hello(self, msg):
        super(AbstractRelayConnection, self).msg_hello(msg)
        self.node.on_relay_connection_ready(self.CONNECTION_TYPE)

    def msg_broadcast(self, msg):
        """
        Handle broadcast message receive from bloXroute.
        This is typically an encrypted block.
        """
        if ConnectionType.RELAY_BLOCK in self.CONNECTION_TYPE:
            self.node.block_processing_service.process_block_broadcast(msg, self)
        else:
            self.log_error(log_messages.UNEXPECTED_BLOCK_ON_NON_RELAY_CONN, msg)

    def msg_key(self, msg):
        """
        Handles key message receive from bloXroute.
        Looks for the encrypted block and decrypts; otherwise stores for later.
        """
        if ConnectionType.RELAY_BLOCK in self.CONNECTION_TYPE:
            self.node.block_processing_service.process_block_key(msg, self)
        else:
            self.log_error(log_messages.UNEXPECTED_KEY_MESSAGE, msg)

    def msg_tx(self, msg):
        """
        Handle transactions receive from bloXroute network.
        """

        start_time = time.time()

        if ConnectionType.RELAY_TRANSACTION not in self.CONNECTION_TYPE:
            self.log_error(log_messages.UNEXPECTED_TX_MESSAGE, msg)
            return

        tx_service = self.node.get_tx_service()

        tx_hash = msg.tx_hash()
        short_id = msg.short_id()
        tx_contents = msg.tx_val()
        is_compact = msg.is_compact()
        network_num = msg.network_num()
        attempt_recovery = False

        ext_start_time = time.time()

        processing_result = tx_service.process_gateway_transaction_from_bdn(tx_hash, short_id, tx_contents, is_compact)

        ext_end_time = time.time()

        if processing_result.ignore_seen:
            gateway_transaction_stats_service.log_duplicate_transaction_from_relay()
            tx_stats.add_tx_by_hash_event(
                tx_hash,
                TransactionStatEventType.TX_RECEIVED_BY_GATEWAY_FROM_PEER_IGNORE_SEEN,
                network_num,
                short_id,
                peers=[self],
                is_compact_transaction=False
            )
            self.log_trace("Transaction has already been seen: {}", tx_hash)
            return

        if processing_result.existing_short_id:
            gateway_transaction_stats_service.log_duplicate_transaction_from_relay(is_compact)
            tx_stats.add_tx_by_hash_event(
                tx_hash,
                TransactionStatEventType.TX_RECEIVED_BY_GATEWAY_FROM_PEER_IGNORE_SEEN,
                network_num,
                short_id,
                peers=[self],
                is_compact_transaction=is_compact
            )
            return

        tx_stats.add_tx_by_hash_event(
            tx_hash,
            TransactionStatEventType.TX_RECEIVED_BY_GATEWAY_FROM_PEER,
            network_num,
            short_id,
            peers=[self],
            is_compact_transaction=msg.is_compact()
        )
        gateway_transaction_stats_service.log_transaction_from_relay(
            tx_hash,
            short_id is not None,
            msg.is_compact()
        )

        if processing_result.assigned_short_id:
            was_missing = self.node.block_recovery_service.check_missing_sid(
                short_id, RecoveredTxsSource.TXS_RECEIVED_FROM_BDN
            )
            attempt_recovery |= was_missing
            tx_stats.add_tx_by_hash_event(
                tx_hash,
                TransactionStatEventType.TX_SHORT_ID_STORED_BY_GATEWAY,
                network_num,
                short_id,
                was_missing=was_missing
            )
            gateway_transaction_stats_service.log_short_id_assignment_processed()

        if not is_compact and processing_result.existing_contents:
            gateway_transaction_stats_service.log_redundant_transaction_content()

        if processing_result.set_content:
            self.log_trace("Adding hash value to tx service and forwarding it to node")
            gateway_bdn_performance_stats_service.log_tx_from_bdn(
                not self.node.is_gas_price_above_min_network_fee(tx_contents)
            )
            attempt_recovery |= self.node.block_recovery_service.check_missing_tx_hash(
                tx_hash, RecoveredTxsSource.TXS_RECEIVED_FROM_BDN
            )

            self.publish_new_transaction(
                tx_hash, tx_contents, TransactionFlag.LOCAL_REGION in msg.transaction_flag()
            )

            if self.node.has_active_blockchain_peer():
                blockchain_tx_message = self.node.message_converter.bx_tx_to_tx(msg)
                transaction_feed_stats_service.log_new_transaction(tx_hash)

                sent = self.node.broadcast_transactions_to_nodes(blockchain_tx_message, self)
                if sent:
                    tx_stats.add_tx_by_hash_event(
                        tx_hash,
                        TransactionStatEventType.TX_SENT_FROM_GATEWAY_TO_BLOCKCHAIN_NODE,
                        network_num,
                        short_id
                    )
                    gateway_bdn_performance_stats_service.log_tx_sent_to_nodes()
                else:
                    gateway_transaction_stats_service.log_dropped_transaction_from_relay()

        if attempt_recovery:
            self.node.block_processing_service.retry_broadcast_recovered_blocks(self)

        end_time = time.time()

        total_duration_ms = (end_time - start_time) * 1000
        duration_before_ext_ms = (ext_start_time - start_time) * 1000
        duration_ext_ms = (ext_end_time - ext_start_time) * 1000
        duration_after_ext_ms = (end_time - ext_end_time) * 1000

        gateway_transaction_stats_service.log_processed_bdn_transaction(
            total_duration_ms,
            duration_before_ext_ms,
            duration_ext_ms,
            duration_after_ext_ms
        )

        performance_utils.log_operation_duration(
            msg_handling_logger,
            "Process single transaction from BDN",
            start_time,
            gateway_constants.BDN_TX_PROCESSING_TIME_WARNING_THRESHOLD_S,
            connection=self,
            message=msg,
            total_duration_ms=total_duration_ms,
            duration_before_ext_ms=duration_before_ext_ms,
            duration_ext_ms=duration_ext_ms,
            duration_after_ext_ms=duration_after_ext_ms
        )

    def msg_txs(self, msg: TxsMessage, recovered_txs_source: RecoveredTxsSource = RecoveredTxsSource.TXS_RECOVERED):
        transactions = msg.get_txs()
        tx_service = self.node.get_tx_service()
        missing_txs: Set[MissingTransactions] = tx_service.process_txs_message(msg)

        tx_stats.add_txs_by_short_ids_event(
            map(lambda x: x.short_id, transactions),
            TransactionStatEventType.TX_UNKNOWN_SHORT_IDS_REPLY_RECEIVED_BY_GATEWAY_FROM_RELAY,
            network_num=self.node.network_num,
            peers=[self],
            found_tx_hashes=map(lambda x: convert.bytes_to_hex(x.hash.binary),
                                transactions
                                )
        )

        for (short_id, transaction_hash) in missing_txs:
            self.node.block_recovery_service.check_missing_sid(short_id, recovered_txs_source)
            self.node.block_recovery_service.check_missing_tx_hash(transaction_hash, recovered_txs_source)

            tx_stats.add_tx_by_hash_event(
                transaction_hash,
                TransactionStatEventType.TX_UNKNOWN_TRANSACTION_RECEIVED_BY_GATEWAY_FROM_RELAY,
                self.node.network_num,
                short_id,
                peers=[self]
            )

        self.node.block_processing_service.retry_broadcast_recovered_blocks(self)

        for block_awaiting_recovery in self.node.block_recovery_service.get_blocks_awaiting_recovery():
            self.node.block_processing_service.schedule_recovery_retry(block_awaiting_recovery)

    def msg_compressed_block_txs(self, msg: CompressedBlockTxsMessage):
        start_time = time.time()

        self.msg_txs(msg.to_txs_message(), RecoveredTxsSource.COMPRESSED_BLOCK_TXS_RECEIVED)

        block_stats.add_block_event_by_block_hash(
            msg.block_hash(),
            BlockStatEventType.ENC_BLOCK_COMPRESSED_TXS_RECEIVED_BY_GATEWAY,
            network_num=msg.network_num(),
            peers=[self],
            more_info="{}. processing time: {}".format(
                stats_format.connection(self),
                stats_format.timespan(start_time, time.time())
            )
        )

    def msg_block_holding(self, msg):
        """
        Block holding request message handler. Places block on hold and broadcasts message to relay and gateway peers.
        :param msg: Message of type BlockHoldingMessage
        """
        block_hash = msg.block_hash()
        self.node.block_processing_service.place_hold(block_hash, self)

    def msg_disconnect_relay_peer(self, _msg: DisconnectRelayPeerMessage) -> None:
        """
        Drop relay peer request handler. Forces a gateway to drop its relay connection and request a new one
        :return: None
        """
        if self.peer_ip in set([outbound_peer.ip for outbound_peer in self.node.opts.outbound_peers if outbound_peer]):
            self.log_info("Received disconnect request. Not dropping because relay peer is static.")
        else:
            self.log_info("Received disconnect request. Dropping.")
            self.mark_for_close(should_retry=False)

    def log_connection_mem_stats(self) -> None:
        """
        logs the connection's memory stats
        """
        super(AbstractRelayConnection, self).log_connection_mem_stats()

        class_name = self.__class__.__name__
        if self.node.message_converter is not None:
            hooks.add_obj_mem_stats(
                class_name,
                self.network_num,
                self.node.message_converter,
                "message_converter",
                memory_utils.ObjectSize(
                    "message_converter", memory_utils.get_special_size(self.node.message_converter).size,
                    is_actual_size=True
                ),
                object_item_count=1,
                object_type=memory_utils.ObjectType.META,
                size_type=memory_utils.SizeType.SPECIAL
            )

    def send_bdn_performance_stats(self, bdn_stats_interval: GatewayBdnPerformanceStatInterval):
        memory_utilization_mb = int(memory_utils.get_app_memory_usage() / constants.BYTE_TO_MB)
        bdn_stats_per_node = bdn_stats_interval.blockchain_node_to_bdn_stats
        if bdn_stats_per_node is None:
            bdn_stats_per_node = {}

        msg_to_send = BdnPerformanceStatsMessage(
            bdn_stats_interval.start_time,
            bdn_stats_interval.end_time,
            memory_utilization_mb,
            bdn_stats_per_node
        )
        self.enqueue_msg(msg_to_send)

    def msg_cleanup(self, msg: AbstractCleanupMessage):
        self.node.block_cleanup_service.process_cleanup_message(msg, self.node)

    def msg_notify(self, msg: NotificationMessage) -> None:
        if msg.notification_code() == NotificationCode.QUOTA_FILL_STATUS:
            args_list = msg.raw_message().split(",")
            entity_type = EntityType(int(args_list[1]))
            quota_level = int(args_list[0])
            if entity_type == EntityType.TRANSACTION and self.node.quota_level != quota_level:
                self.node.quota_level = quota_level
                self.node.alarm_queue.register_approx_alarm(
                    2 * constants.MIN_SLEEP_TIMEOUT, constants.MIN_SLEEP_TIMEOUT, status_log.update_alarm_callback,
                    self.node.connection_pool, self.node.opts.use_extensions, self.node.opts.source_version,
                    self.node.opts.external_ip, self.node.opts.continent, self.node.opts.country,
                    self.node.opts.should_update_source_version, self.node.blockchain_peers, self.node.account_id,
                    self.node.quota_level
                )
        elif (
            msg.notification_code() == NotificationCode.ASSIGNING_SHORT_IDS
            or msg.notification_code() == NotificationCode.NOT_ASSIGNING_SHORT_IDS
        ):
            is_assigning = msg.notification_code() == NotificationCode.ASSIGNING_SHORT_IDS
            peer_model = self.peer_model
            if peer_model is not None:
                peer_model.assigning_short_ids = is_assigning

        if msg.level() == LogLevel.WARNING or msg.level() == LogLevel.ERROR:
            self.log(msg.level(), log_messages.NOTIFICATION_FROM_RELAY, msg.formatted_message())
        else:
            self.log(msg.level(), "Notification from Relay: {}", msg.formatted_message())

    def msg_refresh_blockchain_network(self, _msg: RefreshBlockchainNetworkMessage) -> None:
        blockchain_protocol = self.node.opts.blockchain_protocol
        blockchain_network = self.node.opts.blockchain_network
        assert blockchain_protocol is not None
        assert blockchain_network is not None

        self.node.requester.send_threaded_request(
            sdn_http_service.fetch_blockchain_network,
            blockchain_protocol,
            blockchain_network,
            # pyre-fixme[6]: Expected `Optional[Callable[[Future[Any]], Any]]` for 4th parameter `done_callback`
            #  to call `send_threaded_request` but got `BoundMethod[Callable(_process_blockchain_network_from_sdn)
            #  [[Named(self, AbstractRelayConnection), Named(get_blockchain_network_future, Future[Any])], Any],
            #  AbstractRelayConnection]`.
            done_callback=self._process_blockchain_network_from_sdn
        )

    def publish_new_transaction(
        self, tx_hash: Sha256Hash, tx_contents: memoryview, local_region: bool
    ) -> None:
        self.node.feed_manager.publish_to_feed(
            FeedKey(NewTransactionFeed.NAME, network_num=self.network_num),
            RawTransactionFeedEntry(tx_hash, tx_contents, local_region=local_region)
        )

    def on_connection_established(self):
        super(AbstractRelayConnection, self).on_connection_established()

        if self.check_matching_relay_connection_alarm_id is not None:
            self.node.alarm_queue.unregister_alarm(self.check_matching_relay_connection_alarm_id)
            self.check_matching_relay_connection_alarm_id = None

    def mark_for_close(self, should_retry: Optional[bool] = None):
        super(AbstractRelayConnection, self).mark_for_close(should_retry)
        self.cancel_check_matching_relay_connection()

    def cancel_check_matching_relay_connection(self):
        if self.check_matching_relay_connection_alarm_id is not None:
            self.node.alarm_queue.unregister_alarm(self.check_matching_relay_connection_alarm_id)
            self.check_matching_relay_connection_alarm_id = None

    def _check_matching_relay_connection(self):
        self.log_debug("Verifying that matching relay connection has been established.")

        matching_relay_ip = self.peer_ip
        if self.CONNECTION_TYPE == ConnectionType.RELAY_TRANSACTION:
            matching_relay_port = self.peer_port - 1
        else:
            matching_relay_port = self.peer_port + 1

        if self.node.connection_pool.has_connection(matching_relay_ip, matching_relay_port):
            matching_connection = self.node.connection_pool.by_ipport[(matching_relay_ip, matching_relay_port)]
            if matching_connection.is_active():
                self.log_debug("Matching connection check for relay connection {} is successful. "
                               "Matching connection is {}", self, matching_connection)
                return

        self.log_debug("Closing relay connection {} because matching relay connection is not available.", self)
        if self.CONNECTION_TYPE == ConnectionType.RELAY_TRANSACTION:
            self.node.remove_relay_transaction_peer(self.peer_ip, self.peer_port, True)
        else:
            self.node.remove_relay_peer(self.peer_ip, self.peer_port)

    def _process_blockchain_network_from_sdn(self, get_blockchain_network_future: Future):
        try:
            updated_blockchain_network = get_blockchain_network_future.result()
            assert updated_blockchain_network is not None

            self.node.network = updated_blockchain_network
            self.node.opts.blockchain_networks[self.node.network_num] = updated_blockchain_network
            self.node.update_node_settings_from_blockchain_network(updated_blockchain_network)
        except Exception as e:
            self.log_info("Got {} when trying to read blockchain network from sdn", e)



from typing import Dict, Any, Union

from bxcommon.feed.eth.eth_transaction_feed_entry import EthTransactionFeedEntry
from bxcommon.utils.object_hash import Sha256Hash


class EthTransactionNoSignatureFeedEntry(EthTransactionFeedEntry):
    def __init__(self, tx_hash: Sha256Hash, tx_contents: Union[memoryview, Dict[str, Any]], local_region: bool):
        super().__init__(tx_hash, tx_contents, local_region)
        self.tx_contents.pop("v", None)
        self.tx_contents.pop("r", None)
        self.tx_contents.pop("s", None)



from bxgateway.messages.eth.abstract_eth_message import AbstractEthMessage
from bxcommon.utils.blockchain_utils.eth import crypto_utils, rlp_utils, eth_common_constants
from bxcommon.utils.blockchain_utils.eth import eth_common_utils
from bxgateway.eth_exceptions import WrongMACError, InvalidSignatureError


class EthDiscoveryMessage(AbstractEthMessage):

    def __init__(self, msg_bytes, private_key=None, *args, **kwargs):
        super(EthDiscoveryMessage, self).__init__(msg_bytes, *args, **kwargs)

        self._private_key = private_key
        self._public_key = None

    def __repr__(self):
        return "EthDiscoveryMessage<type: {}>".format(self.__class__.__name__)

    def get_public_key(self):
        if not self._is_deserialized:
            self.deserialize()

        return self._public_key

    def serialize(self):
        cmd_id = rlp_utils.str_to_bytes(chr(self.msg_type))
        encoded_payload = self._serialize_rlp_payload()
        signed_data = eth_common_utils.keccak_hash(cmd_id + encoded_payload)
        signature = crypto_utils.sign(signed_data, self._private_key)

        assert len(signature) == eth_common_constants.SIGNATURE_LEN
        mdc = eth_common_utils.keccak_hash(signature + cmd_id + encoded_payload)

        assert len(mdc) == eth_common_constants.MDC_LEN
        msg_bytes = bytearray(mdc + signature + cmd_id + encoded_payload)

        self._set_raw_bytes(msg_bytes)

    def deserialize(self):
        if self._msg_bytes is None or len(self._msg_bytes) < eth_common_constants.MDC_LEN + eth_common_constants.SIGNATURE_LEN:
            raise ValueError("Message bytes empty or too short.")

        mdc = self._memory_view[:eth_common_constants.MDC_LEN].tobytes()
        if mdc != eth_common_utils.keccak_hash(self._memory_view[eth_common_constants.MDC_LEN:].tobytes()):
            raise WrongMACError("Message hash does not match MDC")

        mdc_sig_len = eth_common_constants.MDC_LEN + eth_common_constants.SIGNATURE_LEN
        signature = self._memory_view[eth_common_constants.MDC_LEN:mdc_sig_len].tobytes()

        self._msg_type = rlp_utils.safe_ord(self._memory_view[mdc_sig_len])

        encoded_data = self._memory_view[mdc_sig_len:].tobytes()
        signed_data = eth_common_utils.keccak_hash(encoded_data)
        remote_pubkey = crypto_utils.recover_public_key(signed_data, signature)

        self._public_key = remote_pubkey

        if not crypto_utils.verify_signature(remote_pubkey, signature, signed_data):
            raise InvalidSignatureError("Message signature does not match public key")

        encoded_payload = self._memory_view[mdc_sig_len + eth_common_constants.MSG_TYPE_LEN:].tobytes()

        self._deserialize_rlp_payload(encoded_payload)

        self._is_deserialized = True

    @classmethod
    def unpack(cls, buf):
        pass

    @classmethod
    def validate_payload(cls, buf, unpacked_args):
        pass



from bxcommon import constants
from bxcommon.messages.bloxroute import short_ids_serializer
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxutils.logging.log_level import LogLevel


class GetTxsMessage(AbstractBloxrouteMessage):
    MESSAGE_TYPE = BloxrouteMessageType.GET_TRANSACTIONS

    """
    Message used to request information about services with specified short ids.
    Node needs to reply with TxsWithShortIdsMessage
    """

    def __init__(self, short_ids=None, buf=None) -> None:

        """
        Constructor. Expects list of short ids or message bytes.

        :param short_ids: list of short ids
        :param buf: message bytes
        """

        self._short_ids = None
        if buf is None:
            buf = self._short_ids_to_bytes(short_ids)
            super(GetTxsMessage, self).__init__(self.MESSAGE_TYPE, len(buf) - self.HEADER_LENGTH, buf)
        else:
            if isinstance(buf, str):
                raise TypeError("Buffer can't be string")

            self.buf = buf
            self._memoryview = memoryview(self.buf)
            self._payload_len = None
            self._payload = None

    def log_level(self):
        return LogLevel.DEBUG

    def get_short_ids(self):
        if self._short_ids is None:
            self._parse()

        return self._short_ids

    def get_serialized_short_ids(self):
        return self.rawbytes()[self.HEADER_LENGTH:-constants.CONTROL_FLAGS_LEN]

    def _short_ids_to_bytes(self, short_ids):
        msg_size = (
            constants.STARTING_SEQUENCE_BYTES_LEN
            + constants.BX_HDR_COMMON_OFF
            + short_ids_serializer.get_serialized_length(len(short_ids))
            + constants.CONTROL_FLAGS_LEN
        )

        buf = bytearray(msg_size)
        off = self.HEADER_LENGTH
        short_ids_serializer.serialize_short_ids_to_buffer(short_ids, buf, off)
        return buf

    def _parse(self):
        self._short_ids, _ = short_ids_serializer.deserialize_short_ids_from_buffer(self.buf, self.HEADER_LENGTH)

    def __repr__(self):
        return "GetTxsMessage<num_short_ids: {}>".format(len(self.get_short_ids()))



from typing import TYPE_CHECKING

from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


class GatewayMemoryUsageRpcRequest(AbstractRpcRequest["AbstractGatewayNode"]):
    help = {
        "params": "",
        "description": "instruct gateway node to log detailed memory usage report"
    }

    def validate_params(self) -> None:
        pass

    async def process_request(self) -> JsonRpcResponse:
        node_size = self.node.get_node_memory_size()
        return self.ok(node_size)



import base64
from typing import Union, Optional, Dict

from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats.network_content_event_type import NetworkContent
from bxutils.logging import LogLevel
from bxutils.logging.log_record_type import LogRecordType
from bxutils import logging


class _NetworkStatisticsService:
    def __init__(self) -> None:
        self.name = "NetworkContent"
        self.log_level = LogLevel.STATS
        self.logger_tx = logging.get_logger(LogRecordType.NetworkContentTx)
        self.logger_block = logging.get_logger(LogRecordType.NetworkContentBlock)

    def transactions_log_event(
            self, network_num: int, tx_hash: Sha256Hash, tx_content: Union[bytearray, memoryview],
            more_info: Optional[Dict] = None) -> None:
        self.logger_tx.log(
            self.log_level,
            {
                "data": {
                    "network_num": network_num,
                    "tx_hash": tx_hash.binary.hex(),
                    "tx_content": base64.b64encode(tx_content),  # pyre-ignore
                    "more_info": more_info
                },
                "type": NetworkContent.TRANSACTION_CONTENT
            }
        )

    def block_content_log_event(
            self, network_num: int, block_hash: Sha256Hash, block_content: memoryview, more_info: Optional[Dict] = None
    ) -> None:
        self.logger_block.log(
            self.log_level,
            {
                "data": {
                    "network_num": network_num,
                    "block_hash": block_hash.binary.hex(),
                    "block_content": base64.b64encode(block_content),  # pyre-ignore
                    "more_info": more_info
                },
                "type": NetworkContent.BLOCK_CONTENT
            }
        )


connection_stats = _NetworkStatisticsService()



from typing import Optional, TYPE_CHECKING, Any, Union, Iterator

from dataclasses import dataclass
from bxcommon.utils.object_hash import Sha256Hash


if TYPE_CHECKING:
    # pylint: disable=ungrouped-imports,cyclic-import
    from task_pool_executor import Sha256 as tpe_Sha256
else:
    tpe_Sha256 = Any

TransactionCacheKeyType = Union[Sha256Hash, bytes, bytearray, memoryview, str, tpe_Sha256]


@dataclass()
class TransactionKey:
    transaction_hash: Sha256Hash
    _transaction_cache_key: Optional[TransactionCacheKeyType] = None
    _lazy_transaction_cache_key: Optional[Iterator[TransactionCacheKeyType]] = None

    @property
    def transaction_cache_key(self) -> TransactionCacheKeyType:
        if self._transaction_cache_key is None:
            lazy_transaction_cache_key = self._lazy_transaction_cache_key
            assert lazy_transaction_cache_key is not None
            self._transaction_cache_key = next(lazy_transaction_cache_key)
        transaction_cache_key = self._transaction_cache_key
        assert transaction_cache_key is not None
        return transaction_cache_key

    def __repr__(self):
        if isinstance(self.transaction_hash, Sha256Hash):
            return self.transaction_hash.binary.hex()
        else:
            return str(self.transaction_hash)



from bxcommon.messages.bloxroute.abstract_cleanup_message import AbstractCleanupMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.utils.object_hash import Sha256Hash


class BlockConfirmationMessage(AbstractCleanupMessage):
    MESSAGE_TYPE = BloxrouteMessageType.BLOCK_CONFIRMATION
    """
    Message with short ids for cleanup for cleanup.
    """

    def block_hash(self) -> Sha256Hash:
        return self.message_hash()

    def __repr__(self):
        return "BlockConfirmationMessage <block_hash: {} :{}> <num_sids: {}> <num_tx_hashes: {}>".format(
            self.block_hash(),
            self.network_num(),
            len(self.short_ids()),
            len(self.transaction_hashes())
        )



from bxcommon.messages.bloxroute.ack_message import AckMessage
from bxcommon.messages.bloxroute.bloxroute_message_factory import _BloxrouteMessageFactory
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.block_holding_message import BlockHoldingMessage
from bxcommon.messages.bloxroute.key_message import KeyMessage
from bxgateway.messages.gateway.block_propagation_request import BlockPropagationRequestMessage
from bxgateway.messages.gateway.block_received_message import BlockReceivedMessage
from bxgateway.messages.gateway.blockchain_sync_request_message import BlockchainSyncRequestMessage
from bxgateway.messages.gateway.blockchain_sync_response_message import BlockchainSyncResponseMessage
from bxgateway.messages.gateway.confirmed_block_message import ConfirmedBlockMessage
from bxgateway.messages.gateway.confirmed_tx_message import ConfirmedTxMessage
from bxgateway.messages.gateway.gateway_hello_message import GatewayHelloMessage
from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType
from bxgateway.messages.gateway.request_tx_stream_message import RequestTxStreamMessage


class _GatewayMessageFactory(_BloxrouteMessageFactory):
    _MESSAGE_TYPE_MAPPING = {
        GatewayMessageType.HELLO: GatewayHelloMessage,
        BloxrouteMessageType.ACK: AckMessage,
        GatewayMessageType.BLOCK_RECEIVED: BlockReceivedMessage,
        BloxrouteMessageType.BLOCK_HOLDING: BlockHoldingMessage,
        GatewayMessageType.BLOCK_PROPAGATION_REQUEST: BlockPropagationRequestMessage,
        BloxrouteMessageType.KEY: KeyMessage,
        GatewayMessageType.CONFIRMED_TX: ConfirmedTxMessage,
        GatewayMessageType.REQUEST_TX_STREAM: RequestTxStreamMessage,
        GatewayMessageType.CONFIRMED_BLOCK: ConfirmedBlockMessage,

        # Sync messages are currently unused. See `blockchain_sync_service.py`
        GatewayMessageType.SYNC_REQUEST: BlockchainSyncRequestMessage,
        GatewayMessageType.SYNC_RESPONSE: BlockchainSyncResponseMessage
    }


gateway_message_factory = _GatewayMessageFactory()



import dataclasses
import functools

from dataclasses import dataclass
from collections import defaultdict
from typing import Optional, Union, Dict, Type, Any, TYPE_CHECKING

from prometheus_client import Counter

from bxcommon import constants
from bxcommon.network.network_direction import NetworkDirection
from bxcommon.utils.stats.measurement_type import MeasurementType
from bxcommon.utils.stats.peer_stats import PeerStats
from bxcommon.utils.stats.statistics_service import StatisticsService, StatsIntervalData
from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType

logger = logging.get_logger(LogRecordType.Throughput, __name__)

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode


bytes_received = Counter("bytes_received", "Number of bytes received from all connections")
bytes_sent = Counter("bytes_sent", "Number of bytes sent on all connections")


@dataclass
class ThroughputIntervalData(StatsIntervalData):
    total_in: int = 0
    total_out: int = 0
    peer_to_stats: Dict[str, PeerStats] = dataclasses.field(
        default_factory=functools.partial(defaultdict, PeerStats)
    )


class ThroughputStatistics(StatisticsService[ThroughputIntervalData, "AbstractNode"]):
    def __init__(
        self,
        interval=constants.THROUGHPUT_STATS_INTERVAL_S,
        look_back=constants.THROUGHPUT_STATS_LOOK_BACK,
    ) -> None:
        super(ThroughputStatistics, self).__init__(
            "ThroughputStats", interval, look_back, reset=True, stat_logger=logger
        )

    def get_interval_data_class(self) -> Type[ThroughputIntervalData]:
        return ThroughputIntervalData

    def add_event(
        self,
        direction: NetworkDirection,
        msg_type: str,
        msg_size: int,
        peer_desc: str,
        peer_id: Optional[str] = None,
    ) -> None:
        peer_stats = self.interval_data.peer_to_stats[peer_desc]
        peer_stats.address = peer_desc
        if peer_id is not None:
            peer_stats.peer_id = peer_id

        if direction == NetworkDirection.INBOUND:
            peer_stats.messages_received[msg_type].bytes += msg_size
            peer_stats.messages_received[msg_type].count += 1
            peer_stats.peer_total_received += msg_size
            self.interval_data.total_in += msg_size
            bytes_received.inc(msg_size)
        else:
            peer_stats.messages_sent.bytes += msg_size
            peer_stats.messages_sent.count += 1
            peer_stats.peer_total_sent += msg_size
            self.interval_data.total_out += msg_size
            bytes_sent.inc(msg_size)

    def add_measurement(
        self,
        peer_desc: str,
        measure_type: MeasurementType,
        measure_value: Union[int, float],
        peer_id: Optional[str] = None,
    ) -> None:
        if peer_id is not None:
            peer_stats = self.interval_data.peer_to_stats[peer_desc]
            peer_stats.peer_id = peer_id
        else:
            peer_stats = self.interval_data.peer_to_stats[peer_desc]
        peer_stats.address = peer_desc

        if measure_type is MeasurementType.PING:
            peer_stats.ping_max = max(peer_stats.ping_max, measure_value)
        elif measure_type is MeasurementType.PING_INCOMING:
            peer_stats.ping_incoming_max = max(peer_stats.ping_incoming_max, measure_value)
        elif measure_type is MeasurementType.PING_OUTGOING:
            peer_stats.ping_outgoing_max = max(peer_stats.ping_outgoing_max, measure_value)

        else:
            raise ValueError(f"Unexpected throughput measurement: {measure_type}={measure_value}")

    def get_info(self) -> Dict[str, Any]:
        node = self.node
        assert node is not None

        node_peers = []
        for conn in node.connection_pool:
            node_peers.append(
                {
                    "peer_address": "%s:%d" % (conn.peer_ip, conn.peer_port),
                    "peer_id": conn.peer_id,
                    "output_buffer_length": conn.get_backlog_size(),
                }
            )

        return {
            "node_id": node.opts.node_id,
            "node_type": node.NODE_TYPE,
            "node_address": f"{node.opts.external_ip}:{node.opts.external_port}",
            "node_peers": node_peers,
            "total_bytes_received": self.interval_data.total_in,
            "total_bytes_sent": self.interval_data.total_out,
            "peer_stats": list(self.interval_data.peer_to_stats.values()),
            "start_time": self.interval_data.start_time,
            "end_time": self.interval_data.end_time,
        }


throughput_statistics = ThroughputStatistics()



import sys
from asyncio import BaseProtocol, BaseTransport, Transport
from typing import Optional, TYPE_CHECKING

import typing

from cryptography.x509 import Certificate

from bxcommon import constants
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.network.socket_connection_protocol_py36 import SocketConnectionProtocolPy36
from bxutils import logging
from bxutils.ssl import ssl_certificate_factory
from bxutils.ssl.extensions import extensions_factory

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode

logger = logging.get_logger(__name__)


class BaseSocketConnectionProtocol(BaseProtocol):
    _delegate_protocol: Optional[AbstractSocketConnectionProtocol]

    def __init__(
        self,
        node: "AbstractNode",
        endpoint: Optional[IpEndpoint] = None,
        is_ssl: bool = True,
    ):
        self._node = node
        self._endpoint = endpoint
        self._is_server = endpoint is None
        self._delegate_protocol = None
        self.is_ssl = is_ssl

    def __getattr__(self, item):
        delegate_protocol = self._delegate_protocol
        if delegate_protocol is not None:
            result = getattr(delegate_protocol, item)
            return result
        return None

    def connection_made(self, transport: BaseTransport) -> None:
        transport = typing.cast(Transport, transport)

        if self._endpoint is None:
            self._endpoint = IpEndpoint(*transport.get_extra_info("peername"))

        endpoint = self._endpoint
        assert endpoint is not None

        if self._is_server:
            try:
                cert = self.get_peer_certificate(transport)
                key = extensions_factory.get_node_id(cert)
                if key is None:
                    key = endpoint.ip_address
            except TypeError:
                key = endpoint.ip_address
            attempts = self._node.report_connection_attempt(key)
            if attempts >= constants.MAX_HIGH_RECONNECT_ATTEMPTS_ALLOWED:
                logger.debug(
                    "Rejecting connection attempt from {} / {}. Too many attempts: {}",
                    self._endpoint,
                    key,
                    attempts
                )
                # transport.abort()
                # return

        if sys.version.startswith("3.6."):
            protocol_cls = SocketConnectionProtocolPy36
        else:
            from bxcommon.network.socket_connection_protocol import SocketConnectionProtocol
            protocol_cls = SocketConnectionProtocol

        if self._is_server:
            endpoint = None
        delegate_protocol = protocol_cls(self._node, endpoint, self.is_ssl)
        delegate_protocol.connection_made(transport)
        self._delegate_protocol = delegate_protocol

    def connection_lost(self, exc: Optional[Exception]) -> None:
        delegate_protocol = self._delegate_protocol
        if delegate_protocol is not None:
            delegate_protocol.connection_lost(exc)
        self._delegate_protocol = None

    def pause_writing(self) -> None:
        delegate_protocol = self._delegate_protocol
        if delegate_protocol is not None:
            delegate_protocol.pause_writing()

    def resume_writing(self) -> None:
        delegate_protocol = self._delegate_protocol
        if delegate_protocol is not None:
            delegate_protocol.resume_writing()

    def pause_reading(self) -> None:
        delegate_protocol = self._delegate_protocol
        if delegate_protocol is not None:
            delegate_protocol.pause_reading()

    def resume_reading(self) -> None:
        delegate_protocol = self._delegate_protocol
        if delegate_protocol is not None:
            delegate_protocol.resume_reading()

    def get_buffer(self, sizehint: int):
        delegate_protocol = self._delegate_protocol
        if delegate_protocol is not None:
            # pyre-ignore ok, only buffered delegate will have this
            return delegate_protocol.get_buffer(sizehint)

        return bytearray(sizehint)

    def buffer_updated(self, nbytes: int):
        delegate_protocol = self._delegate_protocol
        if delegate_protocol is not None:
            # pyre-ignore ok, only buffered delegate will have this
            delegate_protocol.buffer_updated(nbytes)

    def data_received(self, data: bytes):
        delegate_protocol = self._delegate_protocol
        if delegate_protocol is not None:
            # pyre-ignore ok, only unbuffered delegate will have this
            delegate_protocol.data_received(data)

    def get_peer_certificate(self, transport: Transport) -> Certificate:
        assert transport is not None, "Connection is broken!"
        try:
            return ssl_certificate_factory.get_transport_cert(transport)
        except ValueError as e:
            raise TypeError("Socket is not SSL type!") from e



import struct
import time
from typing import List, Optional, Tuple, Set

from bxcommon import constants
from bxcommon.messages.bloxroute import txs_serializer
from bxcommon.messages.bloxroute.tx_service_sync_txs_message import TxServiceSyncTxsMessage
from bxcommon.messages.bloxroute.txs_serializer import TxContentShortIds
from bxcommon.services.transaction_service import TransactionService, TransactionCacheKeyType
from bxcommon.utils import crypto
from bxcommon.utils.object_hash import Sha256Hash
from bxutils import logging

logger = logging.get_logger(__name__)


def create_txs_service_msg(
    transaction_service: TransactionService,
    tx_service_snap: List[Sha256Hash],
    sync_tx_content: bool = True
) -> List[TxContentShortIds]:
    task_start = time.time()
    txs_content_short_ids: List[TxContentShortIds] = []
    txs_msg_len = 0
    while tx_service_snap:
        transaction_key = transaction_service.get_transaction_key(tx_service_snap.pop())
        short_ids = list(transaction_service.get_short_ids_by_key(transaction_key))
        if sync_tx_content:
            tx_content = transaction_service.get_transaction_by_key(transaction_key)
        else:
            tx_content = bytearray(0)
        # TODO: evaluate short id quota type flag value
        short_id_flags = [transaction_service.get_short_id_transaction_type(short_id) for short_id in short_ids]
        tx_content_short_ids: TxContentShortIds = TxContentShortIds(
            transaction_key.transaction_hash,
            tx_content,
            short_ids,
            short_id_flags
        )

        txs_msg_len += txs_serializer.get_serialized_tx_content_short_ids_bytes_len(tx_content_short_ids)

        txs_content_short_ids.append(tx_content_short_ids)
        if txs_msg_len >= constants.TXS_MSG_SIZE or time.time() - task_start > constants.TXS_SYNC_TASK_DURATION:
            break
    return txs_content_short_ids


# pylint: disable=protected-access
def create_txs_service_msg_from_time(
    transaction_service: TransactionService,
    start_time: float = 0,
    sync_tx_content: bool = True,
    snapshot_cache_keys: Optional[Set[TransactionCacheKeyType]] = None
) -> Tuple[List[TxContentShortIds], float, bool, Set[TransactionCacheKeyType]]:
    task_start = time.time()
    txs_content_short_ids: List[TxContentShortIds] = []
    txs_msg_len = 0
    if snapshot_cache_keys is None:
        snapshot_cache_keys = set()
    done = False
    timestamp = start_time
    expire_short_ids = []
    for short_id, timestamp in transaction_service._tx_assignment_expire_queue.queue.items():
        if timestamp > start_time:
            cache_key = transaction_service._short_id_to_tx_cache_key.get(short_id, None)
            if cache_key is not None:
                transaction_key = transaction_service.get_transaction_key(None, cache_key)
                if cache_key not in snapshot_cache_keys:
                    snapshot_cache_keys.add(transaction_key.transaction_cache_key)
                    short_ids = list(
                        transaction_service._tx_cache_key_to_short_ids[transaction_key.transaction_cache_key]
                    )
                    if sync_tx_content and transaction_service.has_transaction_contents_by_key(transaction_key):
                        tx_content = transaction_service._tx_cache_key_to_contents[
                            transaction_key.transaction_cache_key
                        ]
                    else:
                        tx_content = bytearray(0)
                    short_id_flags = [
                        transaction_service.get_short_id_transaction_type(short_id) for short_id in short_ids
                    ]
                    tx_content_short_ids: TxContentShortIds = TxContentShortIds(
                        transaction_key.transaction_hash,
                        tx_content,
                        short_ids,
                        short_id_flags
                    )
                    txs_msg_len += txs_serializer.get_serialized_tx_content_short_ids_bytes_len(tx_content_short_ids)
                    txs_content_short_ids.append(tx_content_short_ids)
                    if txs_msg_len >= constants.TXS_MSG_SIZE or (
                        time.time() - task_start > constants.TXS_SYNC_TASK_DURATION):
                        break
            else:
                expire_short_ids.append(short_id)
    else:
        done = True
    for short_id in expire_short_ids:
        transaction_service._tx_assignment_expire_queue.remove(short_id)
    return txs_content_short_ids, timestamp, done, snapshot_cache_keys


def create_txs_service_msg_from_buffer(
    transaction_service: TransactionService,
    txs_buffer: memoryview,
    start_offset: int
) -> Tuple[TxServiceSyncTxsMessage, int, int, bool]:
    max_task_complete_time = time.time() + constants.GATEWAY_SYNC_BUILD_MESSAGE_THRESHOLD_S
    current_pos = start_offset
    complete_buffer = False
    tx_count = 0

    if start_offset == 0 and len(txs_buffer) <= constants.GATEWAY_SYNC_MAX_MESSAGE_SIZE_BYTES:
        complete_buffer = True
        tx_count, = struct.unpack_from("<L", txs_buffer)
        start_offset = constants.UL_INT_SIZE_IN_BYTES
        current_pos = len(txs_buffer)
    else:
        if start_offset == 0:
            start_offset = current_pos = constants.UL_INT_SIZE_IN_BYTES

        while not complete_buffer and \
            time.time() < max_task_complete_time and \
            current_pos - start_offset <= constants.GATEWAY_SYNC_MAX_MESSAGE_SIZE_BYTES:

            current_pos += crypto.SHA256_HASH_LEN
            content_len, = struct.unpack_from("<L", txs_buffer, current_pos)

            short_ids_offset = (
                current_pos
                + constants.UL_INT_SIZE_IN_BYTES
                + constants.UL_INT_SIZE_IN_BYTES
                + content_len
            )
            short_ids, = struct.unpack_from("<H", txs_buffer, short_ids_offset)
            current_pos = short_ids_offset + constants.UL_SHORT_SIZE_IN_BYTES + (
                (constants.UL_INT_SIZE_IN_BYTES + constants.TRANSACTION_FLAG_LEN) * short_ids
            )
            tx_count += 1
            if current_pos >= len(txs_buffer):
                complete_buffer = True

    return TxServiceSyncTxsMessage(
        transaction_service.network_num,
        txs_buffer=txs_buffer[start_offset:current_pos],
        tx_count=tx_count,
    ), tx_count, current_pos, complete_buffer



class NodeStartArgs:
    EXTERNAL_PORT = "external_port"
    NON_SSL_PORT = "non_ssl_port"
    SDN_ROOT_URL = "sdn_root_url"
    CA_CERT_URL = "ca_cert_url"
    PRIVATE_SSL_BASE_URL = "private_ssl_base_url"



import struct
from typing import Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxutils.logging.log_level import LogLevel


class GetTxContentsMessage(AbstractBloxrouteMessage):
    MESSAGE_TYPE = BloxrouteMessageType.GET_TX_CONTENTS

    """
    Message used to request transaction contents.
    Node needs to reply with TxContentsMessage
    """

    def __init__(self, network_num: Optional[int] = None, short_id: Optional[int] = None,
                 buf: Optional[bytearray] = None):
        """
        Constructor. Expects network num and short id, or message bytes.

        :param network_num: blockchain network number
        :param short_id: short id
        :param buf: message bytes
        """
        msg_size = self.HEADER_LENGTH + constants.NETWORK_NUM_LEN + constants.SID_LEN + constants.CONTROL_FLAGS_LEN

        self._network_num = None
        self._short_id = None
        if buf is None:
            buf = bytearray(msg_size)
            off = self.HEADER_LENGTH

            struct.pack_into("<L", buf, off, network_num)
            off += constants.NETWORK_NUM_LEN
            struct.pack_into("<L", buf, off, short_id)
            off += constants.SID_LEN

            super(GetTxContentsMessage, self).__init__(self.MESSAGE_TYPE, msg_size - self.HEADER_LENGTH, buf)
        else:
            if isinstance(buf, str):
                raise TypeError("Buffer can't be string")

            self.buf = buf
            self._memoryview = memoryview(self.buf)
            self._payload_len = None
            self._payload = None

    def log_level(self):
        return LogLevel.DEBUG

    def network_num(self) -> int:
        if self._network_num is None:
            self._parse()

        assert self._network_num
        # pyre-fixme[7]: Expected `int` but got `None`.
        return self._network_num

    def get_short_id(self) -> int:
        if self._short_id is None:
            self._parse()

        assert self._short_id
        # pyre-fixme[7]: Expected `int` but got `None`.
        return self._short_id

    def _parse(self) -> None:
        off = self.HEADER_LENGTH

        self._network_num, = struct.unpack_from("<L", self.buf, off)
        off += constants.NETWORK_NUM_LEN

        self._short_id, = struct.unpack_from("<L", self.buf, off)
        off += constants.SID_LEN

    def __repr__(self):
        return f"GetTxContentsMessage<network_num: {self.network_num()}, short_id: {self.get_short_id()}>"



from abc import abstractmethod

from bxutils import logging

from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import Sha256Hash

from bxgateway.services.abstract_block_cleanup_service import AbstractBlockCleanupService
from bxgateway.services.btc.btc_block_queuing_service import BtcBlockQueuingService
from bxgateway.messages.btc.block_btc_message import BlockBtcMessage
from bxgateway.messages.btc.inventory_btc_message import GetDataBtcMessage, InventoryType


logger = logging.get_logger(__name__)


class AbstractBtcBlockCleanupService(AbstractBlockCleanupService):
    """
    Service for managing block cleanup.
    """

    # pyre-fixme[11]: Annotation `BtcGatewayNode` is not defined as a type.
    def __init__(self, node: "BtcGatewayNode", network_num: int):
        """
        Constructor
        :param node: reference to node object
        :param network_num: network number
        """

        super(AbstractBtcBlockCleanupService, self).__init__(node=node, network_num=network_num)

    def clean_block_transactions_from_block_queue(
            self,
            block_hash: Sha256Hash,
            block_queuing_service: BtcBlockQueuingService
    ) -> None:
        if block_hash in block_queuing_service:
            block_msg = self.node.block_queuing_service_manager.get_block_data(block_hash)
            self.node.block_cleanup_service.clean_block_transactions(
                transaction_service=self.node.get_tx_service(),
                block_msg=block_msg
            )
        else:
            logger.debug("block cleanup from queuing service failed, block is no longer tracked {}", block_hash)

    @abstractmethod
    def clean_block_transactions(
            self,
            block_msg: BlockBtcMessage,
            transaction_service: TransactionService
    ) -> None:
        pass

    def _request_block(self, block_hash: Sha256Hash):
        block_request_message = GetDataBtcMessage(
            magic=self.node.opts.blockchain_net_magic,
            inv_vects=[(InventoryType.MSG_BLOCK, block_hash)],
            request_witness_data=False
        )
        logger.trace("Received block cleanup request: {}", block_hash)
        node_conn = self.node.get_any_active_blockchain_connection()
        if node_conn:
            node_conn.enqueue_msg(block_request_message)
        else:
            logger.debug("Request for block '{}' failed. No connection to node.", repr(block_hash))



import os
from typing import Dict, Any
from logging import Formatter, LogRecord
import json
from datetime import datetime
from bxutils.encoding.json_encoder import EnhancedJSONEncoder
from bxutils import constants

BUILT_IN_ATTRS = {
    "timestamp",
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
}

EXCLUDE_FROM_PLAIN_FORMATTING = {
    "category"
}


class AbstractFormatter(Formatter):
    FORMATTERS = {
        "%": lambda msg, args: str(msg) % args,
        "{": lambda msg, args: str(msg).format(*args)
    }

    def __init__(self, fmt=None, datefmt=None, style="{") -> None:
        super().__init__(fmt=fmt, datefmt=datefmt)
        self._formatter = self.FORMATTERS[style]

    NO_INSTANCE: str = "[Unassigned]"
    instance: str = NO_INSTANCE

    def _handle_args(self, record) -> str:
        try:
            if record.args and isinstance(record.args[0], str) and record.args[0] == constants.HAS_PREFIX:
                prefix = record.args[1]
                r_args = record.args[2:]
                return " ".join([prefix, self._formatter(record.msg, r_args)])
        except KeyError:
            pass
        return self._formatter(record.msg, record.args)


class JSONFormatter(AbstractFormatter):

    def format(self, record):
        return json.dumps(self._format_json(record), cls=EnhancedJSONEncoder)

    def _format_json(self, record: LogRecord,) -> Dict[Any, Any]:
        log_record = {
            "timestamp": datetime.fromtimestamp(record.created),
            "level": record.levelname,
            "name": record.name,
            "pid": os.getpid()
        }
        log_record.update({k: v for k, v in record.__dict__.items() if k not in BUILT_IN_ATTRS})

        if record.args:
            # There has to be a better way to do this...
            log_record["msg"] = self._handle_args(record)
        else:
            log_record["msg"] = record.msg
        if record.exc_info:
            # pyre-fixme[6]: Expected `Union[Tuple[None, None, None],
            #  Tuple[typing.Type[typing.Any], BaseException,
            #  Optional[types.TracebackType]]]` for 1st param but got
            #  `Optional[typing.Union[Tuple[None, None, None],
            #  Tuple[typing.Type[typing.Any], BaseException,
            #  Optional[types.TracebackType]]]]`.
            log_record["exc_info"] = self.formatException(record.exc_info)
        if self.instance != self.NO_INSTANCE:
            log_record["instance"] = self.instance

        return log_record


class FluentJSONFormatter(JSONFormatter):
    # TODO: check if there is a correct way to annotate this
    def format(self, record):
        return EnhancedJSONEncoder().as_dict(self._format_json(record))


class CustomFormatter(AbstractFormatter):
    encoder = EnhancedJSONEncoder()

    def format(self, record) -> str:
        log_record = {k: v for k, v in record.__dict__.items() if
                      k not in BUILT_IN_ATTRS and k not in EXCLUDE_FROM_PLAIN_FORMATTING}
        if record.args and not hasattr(record.msg, "__dict__"):
            record.msg = self._handle_args(record)
            record.args = ()

        record.msg = "{}{}".format(self.encoder.encode(record.msg),
                                   ",".join({" {}={}".format(k, self.encoder.encode(v)) for (k, v)
                                             in log_record.items() if k != "msg"}))
        if self.instance != self.NO_INSTANCE:
            record.instance = self.instance
        return super(CustomFormatter, self).format(record)

    def formatTime(self, record, datefmt=None) -> str:
        current_time = datetime.fromtimestamp(record.created)
        if datefmt:
            s = current_time.astimezone().strftime(datefmt)
        else:
            s = current_time.isoformat()
        return s



from bxcommon import constants
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.utils.message_buffer_builder import PayloadElement, PayloadBlock


class VersionMessage(AbstractBloxrouteMessage):
    """
    Bloxroute message that contains version info.
    """

    BASE_LENGTH = (
        constants.STARTING_SEQUENCE_BYTES_LEN
        + constants.BX_HDR_COMMON_OFF
        + constants.VERSION_NUM_LEN
        + constants.NETWORK_NUM_LEN
    )
    VERSION_MESSAGE_BLOCK = PayloadBlock(
        AbstractBloxrouteMessage.HEADER_LENGTH,
        "VersionMessage",
        0,
        PayloadElement(structure="<L", name="protocol_version"),
        PayloadElement(structure="<L", name="network_num")
    )
    VERSION_MESSAGE_LENGTH = VERSION_MESSAGE_BLOCK.size + constants.CONTROL_FLAGS_LEN

    def __init__(self, msg_type, payload_len, protocol_version, network_num, buf) -> None:
        if protocol_version is not None and network_num is not None:
            if len(buf) < self.BASE_LENGTH:
                raise ValueError("Version message is not long enough.")
            buf = self.VERSION_MESSAGE_BLOCK.build(buf, protocol_version=protocol_version, network_num=network_num)

        self._protocol_version = None
        self._network_num = None
        super(VersionMessage, self).__init__(msg_type, payload_len, buf)

    def __unpack(self):
        contents = self.VERSION_MESSAGE_BLOCK.read(self._memoryview)
        self._protocol_version = contents.get("protocol_version")
        self._network_num = contents.get("network_num")

    def protocol_version(self):
        if self._protocol_version is None:
            self.__unpack()
        return self._protocol_version

    def network_num(self):
        if self._network_num is None:
            self.__unpack()
        return self._network_num



import json
from typing import Any

from bxutils.encoding.json_encoder import EnhancedJSONEncoder


def serialize(obj: Any) -> str:
    """
    Serializes object into a string JSON
    DEPRECATED use json_encoder.to_json directly

    :param obj: object to serialize
    :return: JSON string
    """
    return json.dumps(obj, cls=EnhancedJSONEncoder)



#
# Copyright (C) 2017, bloXroute Labs, All rights reserved.
# See the file COPYING for details.
#
# Exceptions
#
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_connection import AbstractConnection


class ParseError(Exception):
    def __init__(self, msg) -> None:
        super(ParseError, self).__init__(msg)

        self.msg = msg


class UnrecognizedCommandError(ParseError):
    def __init__(self, msg, raw_data) -> None:
        super(UnrecognizedCommandError, self).__init__(msg)

        self.raw_data = raw_data


class PayloadLenError(ParseError):
    pass


class UnauthorizedMessageError(ParseError):
    pass


class ChecksumError(ParseError):
    def __init__(self, msg, raw_data) -> None:
        super(ChecksumError, self).__init__(msg)

        self.raw_data = raw_data


class HighMemoryError(Exception):
    pass


class TerminationError(Exception):
    pass


class DecryptionError(Exception):
    pass


class FeedSubscriptionTimeoutError(Exception):
    pass


class ConnectionStateError(Exception):
    def __init__(self, msg: str, conn: "AbstractConnection") -> None:
        super().__init__(msg)
        self.msg = msg
        self.conn = conn



from typing import cast

from bxcommon.messages.abstract_internal_message import AbstractInternalMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.messages.bloxroute.v17.tx_message_v17 import TxMessageV17
from bxcommon.messages.versioning.abstract_message_converter import AbstractMessageConverter
from bxcommon.models.transaction_flag import TransactionFlag


class _TxMessageConverterV17(AbstractMessageConverter):
    _MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.TRANSACTION: TxMessageV17
    }

    _MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.TRANSACTION: TxMessage
    }

    UNKNOWN_TRANSACTION_FLAGS = (
        TransactionFlag.LOCAL_REGION
        | TransactionFlag.PRIVATE_TX
        | TransactionFlag.TBD_2
        | TransactionFlag.TBD_3
        | TransactionFlag.TBD_4
        | TransactionFlag.TBD_5
    )
    INVERTED_UNKNOWN_TRANSACTION_FLAGS = (
        ~TransactionFlag.LOCAL_REGION
        & ~TransactionFlag.PRIVATE_TX
        & ~TransactionFlag.TBD_2
        & ~TransactionFlag.TBD_3
        & ~TransactionFlag.TBD_4
        & ~TransactionFlag.TBD_5
    )

    def convert_from_older_version(
        self, msg: AbstractInternalMessage
    ) -> AbstractInternalMessage:
        msg_type = msg.MESSAGE_TYPE

        if msg_type not in self._MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING:
            raise ValueError(
                f"Tried to convert unexpected new "
                f"message type to v17: {msg_type}"
            )

        msg = cast(TxMessageV17, msg)

        tx_hash = msg.tx_hash()
        network_num = msg.network_num()
        source_id = msg.source_id()
        short_id = msg.short_id()
        tx_val = msg.tx_val()
        transaction_flag = msg.transaction_flag()
        ts = msg.timestamp()

        return TxMessage(
            message_hash=tx_hash,
            network_num=network_num,
            source_id=source_id,
            short_id=short_id,
            tx_val=tx_val,
            transaction_flag=transaction_flag,
            timestamp=ts
        )

    def convert_to_older_version(
        self, msg: AbstractInternalMessage
    ) -> AbstractInternalMessage:
        msg_type = msg.MESSAGE_TYPE

        if msg_type != TxMessage.MESSAGE_TYPE:
            raise ValueError(
                f"Tried to convert unexpected new "
                f"message type to v17: {msg_type}"
            )

        msg = cast(TxMessage, msg)
        transaction_flag = msg.transaction_flag()

        if self.UNKNOWN_TRANSACTION_FLAGS & transaction_flag:
            transaction_flag &= self.INVERTED_UNKNOWN_TRANSACTION_FLAGS

        return TxMessageV17(
            msg.message_hash(),
            msg.network_num(),
            msg.source_id(),
            msg.short_id(),
            msg.tx_val(),
            transaction_flag,
            int(msg.timestamp())
        )

    def convert_first_bytes_to_older_version(
        self, first_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_first_bytes_from_older_version(
        self, first_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_last_bytes_to_older_version(
        self, last_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_last_bytes_from_older_version(
        self, last_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def get_message_size_change_to_older_version(self) -> int:
        return 0

    def get_message_size_change_from_older_version(self) -> int:
        return 0


tx_message_converter_v17 = _TxMessageConverterV17()



# pylint: disable=global-statement

import configparser
import json
import multiprocessing
import os
import re
import shutil
import sys
from os import path
from pathlib import Path
from typing import Dict, Any

from bxcommon import constants
from bxcommon.messages.bloxroute.bloxroute_version_manager import bloxroute_version_manager
from bxutils import logging

logger = logging.get_logger(__name__)

_working_directory = ""
_data_directory = ""


def set_working_directory(working_directory: str):
    """
    Sets working directory of python application
    :param working_directory: full name of working directory
    """
    global _working_directory
    _working_directory = working_directory


def set_data_directory(data_directory: str):
    """
    Sets the directory where configuration, state and log files are being stored
    :param data_directory: full name of data directory
    """
    global _data_directory
    _data_directory = data_directory
    try:
        os.makedirs(data_directory, exist_ok=True)

    # pylint: disable=broad-except
    except Exception as e:
        logger.fatal(
            "Unable to create data directory '{}'. Error: {}. Exiting.",
            data_directory,
            e
        )
        sys.exit(1)


def get_default_data_path():
    """
    Returns default data directory for configuration, state and log files
    :return: default data directory
    """

    if sys.platform == constants.PLATFORM_MAC:
        return path.join(get_home_path(), "Library", "bloXroute")
    else:
        return path.join(get_home_path(), "bloxroute")


def get_home_path():
    """
    Returns user home directory
    :return: home directory
    """
    return str(Path.home())


def get_relative_file(file_path: str) -> str:
    return os.path.join(_working_directory, file_path)


def get_data_file(file_path: str) -> str:
    return os.path.join(_data_directory, file_path)


def log_pid(file_name):
    """
    Capture process id for easy termination in multi-thread scenarios.
    """
    with open(file_name, "w") as writable_file:
        writable_file.write(str(os.getpid()))


def get_env_default(key: str) -> str:
    """
    Read the default values for given key from the config file for a specific environment. The environment is stored
    in the hosts environment variables as `BLXR_ENV=local` for example. If the environment isn't set or doesn't exist
    in the config file, the global default values are returned

    :param key: the key to get the env default values for
    :return: the env default value
    """
    config = configparser.ConfigParser()
    config_file_path = get_relative_file(constants.NODE_CONFIG_FILE)
    config.read(config_file_path)

    # If the environment var itself does not exist, return the global default value for the key
    if constants.BLXR_ENV_VAR not in os.environ:
        return config.defaults()[key]

    environment = os.environ[constants.BLXR_ENV_VAR].lower()

    # If the given environment var does not have an entry in the config file, return the global default
    if environment in config.sections():
        return config.get(environment, key)
    else:
        return config.defaults()[key]


def get_thread_pool_parallelism_degree(parallelism_degree_str: str) -> int:
    parallelism_degree = int(parallelism_degree_str)
    if parallelism_degree <= 0:
        parallelism_degree = 1
    return min(
        parallelism_degree,
        max(multiprocessing.cpu_count() - 1, 1)
    )


def is_valid_version(full_version: str) -> bool:
    """
    check if version number in template: {int}.{int}.{int}.{int} and version type is dev/ci/v
    :param full_version: {version_type}{version_number}
    :return: if the version is valid
    """
    index = re.search(r"\d", full_version)
    if index is None:
        raise ValueError(f"Version is incorrectly formated: {full_version}")
    index = index.start()

    version_number = full_version[index:]
    version_type = full_version[:index]
    return (version_number.count(".") == 3 and all(str(x).isdigit() for x in version_number.split("."))) and \
           (version_type in constants.VERSION_TYPE_LIST)


def read_manifest(manifest_path: str) -> Dict[str, str]:
    """
    Reads manifest file, ensuring that manifest contains all expected properties.
    :param manifest_path:
    :return: manifest
    """
    try:
        with open(manifest_path, "r") as json_file:
            manifest = json.load(json_file)
            version = manifest[constants.MANIFEST_SOURCE_VERSION]

            if not is_valid_version(version):
                raise ValueError(f"Invalid version: {version}")

            if not all(params in manifest for params in constants.REQUIRED_PARAMS_IN_MANIFEST):
                missing_params_str = ", ".join(
                    [item for item in constants.REQUIRED_PARAMS_IN_MANIFEST if item not in manifest])
                raise ValueError(f"Missing attributes: {missing_params_str}")

            return manifest
    except Exception as e:
        raise Exception(f"Unexpected error while reading manifest file at: {manifest_path}. Error: {e}")


def append_manifest_args(opts_dict: Dict[Any, Any]):
    manifest_path = get_relative_file(constants.MANIFEST_PATH)
    manifest = read_manifest(manifest_path)
    opts_dict.update(manifest)
    opts_dict.update({constants.PROTOCOL_VERSION: bloxroute_version_manager.CURRENT_PROTOCOL_VERSION})


def init_file_in_data_dir(file_name: str):
    relative_file_name = get_relative_file(file_name)
    data_dir_file_name = get_data_file(file_name)

    if os.path.exists(relative_file_name) and not os.path.exists(data_dir_file_name):
        shutil.copyfile(relative_file_name, data_dir_file_name)



from typing import NamedTuple

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.network.transport_layer_protocol import TransportLayerProtocol


class ConnectionPeerInfo(NamedTuple):
    endpoint: IpEndpoint
    connection_type: ConnectionType
    transport_protocol: int = TransportLayerProtocol.TCP

    def __repr__(self) -> str:
        return f"{self.endpoint} [{self.connection_type.name}]"

    def __eq__(self, other: "ConnectionPeerInfo") -> bool:
        return self.endpoint == other.endpoint

    def __hash__(self):
        return hash(self.endpoint)



import datetime

from typing import Optional, List, TYPE_CHECKING
from collections import defaultdict

from bxcommon.utils.stats import stats_format
from bxcommon.utils.stats.stat_event_type_settings import StatEventTypeSettings

if TYPE_CHECKING:
    # pylint: disable=cyclic-import
    from bxcommon.connections.abstract_connection import AbstractConnection


class StatEvent:
    def __init__(
        self,
        event_settings: StatEventTypeSettings,
        event_subject_id: str,
        node_id: str,
        start_date_time: datetime.datetime,
        end_date_time: Optional[datetime.datetime] = None,
        peers: Optional[List["AbstractConnection"]] = None,
        **kwargs
    ):
        self.event_name = event_settings.name
        self.event_logic = event_settings.event_logic
        self.event_subject_id = event_subject_id
        self.node_id = node_id
        self.start_date_time = start_date_time
        self.end_date_time = end_date_time if end_date_time is not None else start_date_time
        self.extra_data = kwargs
        if peers:
            peer_ids = defaultdict(list)
            for peer in peers:
                if peer and peer.peer_id:
                    peer_ids[peer.CONNECTION_TYPE.format_short()].append(peer.peer_id)
            self.extra_data["peer_ids"] = peer_ids
            self.extra_data["peers"] = stats_format.connections(peers)
